@@ -1,5 +1,5 @@
 // 
-// Copyright (c) 2006-2007, Benjamin Kaufmann
+// Copyright (c) 2006-2010, Benjamin Kaufmann
 // 
 // This file is part of Clasp. See http://www.cs.uni-potsdam.de/clasp/ 
 // 
@@ -18,7 +18,7 @@
 // Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 //
 #include "clasp_options.h"
-#include <program_opts/value.h>
+#include "program_opts/value.h"  
 #include <clasp/satelite.h>
 #include <sstream>
 #include <algorithm>
@@ -186,8 +186,10 @@ bool SolverStrategiesWrapper::mapSatElite(const std::string& s, SolverStrategies
 			pre->options.maxIters = v.size()>0 ? v[0] : -1;
 			pre->options.maxOcc   = v.size()>1 ? v[1] : -1;
 			pre->options.maxTime  = v.size()>2 ? v[2] : -1;
+			pre->options.maxFrozen= v.size()>3 && v[3] > 0 ? (v[3]/100.0) : 1.0;
 			i.opts->satPrePro.reset( pre );
 		}
+		i.satPreproDef = false;
 		return true;
 	}
 	return false;
@@ -268,15 +270,17 @@ void GeneralOptions::initOptions(ProgramOptions::OptionGroup& root, ProgramOptio
 		("sat-prepro", storeTo(solverOpts)->parser(&Strategies::mapSatElite)->setImplicit(),
 			"Configure SatELite-like preprocessing\n"
 			"      Default: no\n"
-			"      Valid:   yes, no, <n1[,n2,n3]>\n"
+			"      Valid:   yes, no, <n1[,n2,n3,n4]>\n"
 			"        <n1>: Run for at most <n1> iterations           (-1=run to fixpoint)\n"
 			"        <n2>: Run variable elimination with cutoff <n2> (-1=no cutoff)\n"
 			"        <n3>: Run for at most <n3> seconds              (-1=no time limit)\n"
+			"        <n4>: Disable if <n4>% of vars are frozen       (-1=no time limit)\n"
 			"        yes : Run to fixpoint, no cutoff and no time limit\n","<opts>")
 	;
 	root.addOptions(general);
 	hidden.addOptions()
 		("project-opt", storeTo(config->enumerate.projectOpts), "Additional options for projection as octal digit\n")
+		("dfs-eq", bool_switch(&config->api.eqDfs), "Enable df-order in eq-preprocessing\n")
 	;
 }
 bool GeneralOptions::validateOptions(ProgramOptions::OptionValues& vm, Messages&) {
@@ -463,8 +467,15 @@ bool ClaspOptions::validateOptions(ProgramOptions::OptionValues& vm, Messages& m
 
 void ClaspOptions::addDefaults(std::string& def) {
 	def += "1 --trans-ext=no --eq=5 --sat-prepro=no --rand-watches=yes\n";
-	def += "  --lookahead=no --heuristic=Berkmin --rand-freq=0.0 --rand-prob=no\n";
+  def += "  --lookahead=no --heuristic=Berkmin --rand-freq=0.0 --rand-prob=no\n";
 	def += "  --restarts=100,1.5 --shuffle=0,0 --deletion=3.0,1.1,3.0\n";
 	def += "  --strengthen=all --loops=common --contraction=250\n";
 }
+
+void ClaspOptions::applyDefaults(Input::Format f) {
+	if (f != Input::SMODELS && mode.solverOpts.satPreproDef) {
+		SolverStrategiesWrapper::mapSatElite("20,25,120", mode.solverOpts);
+	}
+}
+
 }
