@@ -33,6 +33,7 @@
 #include <map>
 #include <algorithm>
 #include <iosfwd>
+#include <stdexcept>
 
 namespace Clasp {
 
@@ -167,11 +168,18 @@ struct PreproStats {
 };
 
 
+//! Exception type for signaling an invalid incremental program update.
+class RedefinitionError : public std::logic_error {
+public:
+	explicit RedefinitionError(const std::string& m) : std::logic_error(m) {}
+};
+
+
 //! Interface for defining a logic program.
 /*!
  * Use this class to specify a logic program. Once the program is completly defined,
- * call endProgram to load the logic program into a Solver-object.
- * 
+ * call endProgram() to load the logic program into a Solver-object.
+ *
  */
 class ProgramBuilder {
 public:
@@ -189,9 +197,25 @@ public:
 		mode_transform_dynamic= 6  /**< Heuristically decide whether or not to transform a particular extended rule */
 	};
 
+	/*!
+	 * \name Setup and step control functions
+	 */
+	//@{
+
+	//! Starts the definition of a logic program. 
+	/*!
+	 * This function shall be called exactly once before a new program is defined.
+	 * It discards any previously added program.
+	 *
+	 * \param index The symbol table in which the program builder stores information about atoms
+	 * \param ufs The object to use for unfounded set detection. If 0, unfounded set detection is disabled.
+	 * \note If ufsChecker is not 0 it shall point to a heap-allocated object. The ownership is transferred!
+	 */
+	ProgramBuilder& startProgram(AtomIndex& index, DefaultUnfoundedCheck* ufs);
+
 	//! Sets the mode for handling extended rules.
 	/*!
-	 * The default mode is to handle all extended rules natively.
+	 * The default is to handle all extended rules natively.
 	 */
 	void setExtendedRuleMode(ExtendedRuleMode m) {
 		erMode_ = m;
@@ -208,24 +232,10 @@ public:
 		eqDfs_   = dfs;
 	}
 
-	//! disposes (parts of) the internal representation of the logic program.
-	/*!
-	 * \param forceFullDispose If set to true, the whole program is disposed. Otherwise,
-	 * if updateProgram() was called, only the last incremental step is disposed.
-	 */
-	void disposeProgram(bool forceFullDispose);
-	
-	//! Starts the definition of a logic program
-	/*!
-	 * This function shall be called exactly once before any rules or atoms are added.
-	 * \param index The symbol table in which the program builder stores information about atoms
-	 * \param ufs The object to use for unfounded set detection. If 0, unfounded set detection is disabled.
-	 * \note If ufsChecker is not 0 it shall point to a heap-allocated object. The ownership is transferred!
-	 */
-	ProgramBuilder& startProgram(AtomIndex& index, DefaultUnfoundedCheck* ufs);
-
-	//! Unfreezes a program and starts the next step if program is defined incrementally
+	//! Unfreezes a currently frozen program and starts an incremental step.
 	/*!
+	 * \pre The program is either frozen or at step 0.
+	 *
 	 * If a program is to be defined incrementally, this function must be called
 	 * exactly once for each step before any new rules or atoms are added.
 	 * \note Program update only works correctly under the following assumptions:
@@ -233,57 +243,147 @@ public:
 	 *    - solely defined in step i OR,
 	 *    - marked as frozen in step i and solely defined in step i+k OR,
 	 *    - forced to false by a acompute statement in step 0
-	 *  - Atom ids are monotonically increasing, i.e. atoms defined in step i have lower ids than those defined in step i+k.
-	 *  - Rule bodies are either empty or contain at least one atom defined in the current step
+	 *
+	 * \post The program is no longer frozen and calling program mutating functions is valid again. 
+	 * \throws std::logic_error precondition is violated.
 	 */
 	ProgramBuilder& updateProgram();
-	
-	//! creates a new atom and returns the new atom's id.
+
+	//! Finishes the definition of the logic program (or its current increment).
 	/*!
-	 * \pre The program is not frozen
+	 * Applies program mutating operations issued in the current step and transforms 
+	 * the new program into the solver's internal representation. 
+	 *
+	 * \param solver The solver object in which the representation of the lp should be stored.
+	 * \param finalizeSolver if true, endProgram() calls solver.endAddConstraints()
+	 * \param backprop Enable backpropagation in preprocessing
+	 * \return false if the program is initially conflicting, true otherwise.
+	 *
+	 * \post
+	 *  - If true is returned, the program is considered to be "frozen" and calling 
+	 *    program mutating functions is invalid until the next call to updateProgram().
+	 *  - If false is returned, the state of the object is undefined and startProgram() 
+	 *    and disposeProgram() are the only remaining valid operations. 
+	 *  . 
+	 *
+	 * \note To load the same program into different solvers, 
+	 *       call endProgram for each solver.
+	 */
+	bool endProgram(Solver& solver, bool finalizeSolver, bool backprop = false);
+
+	//! Creates a minimize constraint from the minimize statements contained in the program.
+	/*!
+	 * \pre The program is frozen and not trivially UNSAT, i.e. endProgram() was already called and did not return false.
+	 * \return 0 if the program does not contain minimize statements
+	 */
+	MinimizeConstraint* createMinimize(Solver& s, bool inHeu = false);
+
+	//! Writes the (possibly simplified) program in lparse-format to the given stream.
+	void writeProgram(std::ostream& os);
+
+	//! Disposes (parts of) the internal representation of the logic program.
+	/*!
+	 * \param forceFullDispose If set to true, the whole program is disposed. Otherwise,
+	 *  only the rules (of the current step) are disposed but atoms and any incremental
+	 *  control data remain.
+	 */
+	void disposeProgram(bool forceFullDispose);
+	//@}
+	
+	/*!
+	 * \name Program mutating functions
+	 * 
+	 * Functions in this group shall only be called if the program is currently not 
+	 * frozen. That is, only between the call to startProgram() (resp. updateProgram() if in 
+	 * incremental setting) and endProgram(). A std::logic_error is raised if this precondition is violated. 
+	 *
+	 */
+	//@{
+
+	//! Adds a new atom to the program.
+	/*!
+	 * \return The new atom's id. 
 	 */
 	Var newAtom();
-	//! returns the number of atoms in the current logic program
-	uint32 numAtoms() const  { return (uint32)atoms_.size()-1; }
-	//! returns the number of bodies in the current (slice of the) logic program
-	uint32 numBodies() const { return (uint32)bodies_.size(); }
 
-	//! Sets the name of the Atom with the given id.
+	//! Sets the name of the given atom and adds it to the program's symbol table.
 	/*!
+	 * \pre 
+	 *   - The atom is either not yet known or was added in the current step (atomId >= startAtom()).
+	 *   - The atom was not yet added to the symbol table, i.e. 
+	 *     setAtomName() is called at most once for an atom.
+	 *   . 
 	 * \param atomId The id of the atom for which a name should be set
 	 * \param name The new name of the atom with the given id.
-	 * \pre name != 0 && atomId > 0 && atomId < varMax
-	 * \note if atomId is not yet known, an atom with the given id is implicitly created.
-	 * \pre shall be called at most once per atom
+	 * \note If atomId is not yet known, an atom with the given id is implicitly created. 
+	 *
+	 * \throws RedefinitionError precondition is violated. 
+	 * \throws std::logic_error  program is frozen.
 	 */
 	ProgramBuilder& setAtomName(Var atomId, const char* name);
 
-	//! forces the atom with the given atom id to true/false.
+	//! Forces the atom's truth-value to value. 
 	/*!
+	 * \pre The atom is either not yet known, false, or an atom from the current step.
 	 * \param atomId Id of the Atom for which a truth-value should be set.
 	 * \param pos If true, atom is set to true (forced to be in every answer set). Otherwise
-	 * atom is set to false (not part of any answer set).
+	 *            atom is set to false (not part of any answer set).
+	 * \note If atomId is not yet known, an atom with the given id is implicitly created. 
 	 */
-	ProgramBuilder& setCompute(Var atomId, bool pos);
+	ProgramBuilder& setCompute(Var atomId, bool value);
 
-	//! Marks the atom with the given atom id as frozen.
+	//! Protects an otherwise undefined atom from preprocessing. 
 	/*!
-	 * \note An atom a stays frozen until unfreeze(a) is called.
-	 * \note As long as an atom is frozen it shall not be defined.
+	 * If the atom is defined in this or a previous step, the operation has no effect. 
+	 * \note If atomId is not yet known, an atom with the given id is implicitly created. 
 	 */
 	ProgramBuilder& freeze(Var atomId);
 
-	//! Unfreezes a previously frozen atom.
+	//! Removes any protection from the given atom. 
 	/*!
-	 * \pre atomId must refer to an atom that is currently frozen.
+	 * If the atom is defined in this or a previous step, the operation has no effect. 
+	 * \note
+	 *   - The effect is logically equivalent to adding a rule atomId :- false. 
+	 *   - A call to unfreeze() always overwrites a call to freeze() even if the 
+	 *     latter comes after the former
+	 *   . 
 	 */
 	ProgramBuilder& unfreeze(Var atomId);
+
+	//! Adds the given rule to the program.
+	/*!
+	 * \pre The head of the rule does not contain an atom defined in a 
+	 *      previous incremental step. 
+	 *
+	 * Simplifies the given rule and adds it to the program if it
+	 * is neither tautological (e.g. a :- a) nor contradictory (e.g. a :- b, not b). 
+	 * Atoms in the simplified rule that are not yet known are implicitly created. 
+	 *
+	 * \throws RedefinitionError if the precondition is violated. 
+	 * \note If the head of the simplified rule mentions an atom from a previous step,
+	 *       that atom shall either be frozen or false. In the former case, 
+	 *       unfreeze() is implicitly called. In the latter case, the rule is interpreted 
+	 *       as an integrity constraint. 
+	 */
+	ProgramBuilder& addRule(PrgRule& r);
 	
+	//@}
+
+	/*!
+	 * \name Rule creation functions
+	 * 
+	 * Functions in this group may be used to construct logic program rules.
+	 * The construction of a rule must start with a call to startRule() and
+	 * ends with a call to endRule(). Functions for adding elements to a
+	 * rule shall only be called between calls to startRule()/endRule() and
+	 * only one rule can be under construction at any one time.
+	 */
+	//@{
+
 	//! Starts the construction of a rule.
 	/*! 
 	 * \param t The type of the new rule.
 	 * \param bound The lower bound (resp. min weight) of the rule to be created.
-	 * \pre no rule is active
 	 *
 	 * \note the bound-parameter is only interpreted if the rule to be created is
 	 * either a constraint- or a weight-rule.
@@ -300,7 +400,7 @@ public:
 	//! Sets the bound (resp. min weight) of the currently active rule.
 	/*!
 	 * \param bound The lower bound (resp. min weight) of the rule to be created.
-	 * \pre The active rule is either a constraint or weight rule.
+	 * \pre The rule under construction is either a constraint or weight rule.
 	 */
 	ProgramBuilder& setBound(weight_t bound) { // only valid for CONSTRAINT and WEIGHTRULES
 		rule_.setBound(bound);
@@ -308,10 +408,6 @@ public:
 	}
 
 	//! Adds the atom with the given id as a head to the currently active rule.
-	/*!
-	 * \pre startRule was called && atomId > 0
-	 * \note if atomId is not yet known, an atom with the given id is implicitly created.
-	 */
 	ProgramBuilder& addHead(Var atomId) {
 		assert(atomId > 0);
 		rule_.addHead(atomId);
@@ -320,68 +416,63 @@ public:
 	
 	//! Adds a subgoal to the currently active rule.
 	/*!
+	 * \pre atomId > 0 && weight >= 0
 	 * \param atomId The id of the atom to be added to the rule.
 	 * \param pos true if the atom is positive. Fals otherwise
 	 * \param weight The weight the new predecessor should have in the rule.
-	 * \pre atomId > 0 && weight >= 0
 	 * \note The weight parameter is only used if the active rule is a weight or optimize rule.
-	 * \note if atomId is not yet known, an atom with the given id is implicitly created.
 	 */
 	ProgramBuilder& addToBody(Var atomId, bool pos, weight_t weight = 1) {
 		rule_.addToBody(atomId, pos, weight);
 		return *this;
 	}
 	
-	//! Finishes the creation of the active rule and adds it to the program
-	/*! 
-	 * \pre The program is not frozen
-	 * \note Must be called once for each rule.
+	//! Finishes the construction of the active rule and adds it to the program.
+	/*!
+	 * \see ProgramBuilder::addRule(PrgRule&);
 	 */
 	ProgramBuilder& endRule() {
 		return addRule(rule_);
 	}
-
-	//! Adds the rule r to the program
-	/*!
-	 * \pre The program is not frozen
-	 */
-	ProgramBuilder& addRule(PrgRule& r);
 	
-	//! Finishes the definition of the logic program (or its current increment).
-	/*!
-	 * Call this method to load the program (increment) into the solver
-	 *
-	 * \param solver The solver object in which the optimized representation of the lp should be stored.
-	 * \param finalizeSolver if true, endProgram() calls solver.endAddConstraints()
-	 * \param backprop Enable backpropagation in preprocessing
-	 * \return false if the program is initially conflicting, true otherwise.
-	 *
-	 * \note If endProgram returned false, the state of the ProgramBuilder object is undefined. In that case,
-	 * only startProgram() and disposeProgram() remain valid operations.
-	 *
-	 * \note After endProgram was called, the program is considered "frozen". One shall not add atoms/rules
-	 * to a frozen program. Call ProgramBuilder::updateProgram() to "unfreeze" a program.
-	 * \note To load the same program into different solvers, call endProgram for each solver.
-	 *
-	 */
-	bool endProgram(Solver& solver, bool finalizeSolver, bool backprop = false);
-
-	//! returns true if the program contains at least one minimize statement
-	bool hasMinimize() const { return minimize_ != 0; }
+	//@}
 	
-	//! Creates a minimize constraints from the minimize statements contained in the program
 	/*!
-	 * \pre The program is frozen and not trivially UNSAT, ie. endProgram was already called and did not return false.
-	 * \return 0 if the program does not contain minimize statements
+	 * \name Query functions
+	 *
 	 */
-	MinimizeConstraint* createMinimize(Solver& s, bool inHeu = false);
+	 //@{
 
+	//! Returns true if the program contains any minimize statements.
+	bool   hasMinimize() const { return minimize_ != 0; }
+	//! Returns the number of atoms in the logic program.
+	uint32 numAtoms() const  { return (uint32)atoms_.size()-1; }
+	//! Returns the number of bodies in the current (slice of the) logic program.
+	uint32 numBodies() const { return (uint32)bodies_.size(); }
+	//! Returns the id of the first atom of the current step.
+	Var    startAtom() const { return incData_?incData_->startAtom_:1; }
+
+	//! Returns the internal literal that is associated with the given atom.
 	/*!
-	 * Writes the (possibly simplified) program in lparse-format to the given stream.
-	 * \pre endProgram was called and the program is currently "frozen".
+	 * \pre atomId is a known atom
+	 * \return A literal that is valid in the current solving context. 
+	 * \note Untill endProgram() is called, atoms from the current step are
+	 *       associated with the always-false literal negLit(0).
+	 * \throws std::logic_error if precondition is violated.
 	 */
-	void writeProgram(std::ostream& os);  // Pre: endProgram
+	Literal getLiteral(Var atomId) const;
+	//! Returns a vector of internal literals that, when assumed true, makes all frozen atoms false. 
+	/*!
+	 * \pre The program is currently frozen. 
+	 */
+	void    getAssumptions(LitVec& out) const;
 	PreproStats stats;
+	//@}
+	// FOR TESTING
+	PrgAtomNode*  getAtom(Var atomId) const { return atoms_[atomId]; }
+	PrgBodyNode*  getBody(Var bodyId) const { return bodies_[bodyId]; }
+	Var           getEqAtom(Var a)    const { return getEqNode(atoms_, a);  }
+	const LitVec& getCompute()        const { return compute_; }
 private:
 	ProgramBuilder(const ProgramBuilder&);
 	ProgramBuilder& operator=(const ProgramBuilder&);
@@ -399,6 +490,7 @@ private:
 	// ------------------------------------------------------------------------
 	// Program definition
 	PrgAtomNode*  resize(Var atomId);
+	bool          inCompute(Literal x) const;
 	void          addRuleImpl(const PrgRule& r, const PrgRule::RData& rd);
 	bool          handleNatively(const PrgRule& r, const PrgRule::RData& rd) const;
 	bool          transformNoAux(const PrgRule& r, const PrgRule::RData& rd) const;
@@ -407,8 +499,8 @@ private:
 	void          clearRuleState(const PrgRule& r);
 	Body          findOrCreateBody(const PrgRule& r, const PrgRule::RData& rd);
 	bool          mergeEqAtoms(Var a, Var root);
-	Var           getEqAtom(Var a) const { return getEqNode(atoms_, a); }
 	Var           getEqBody(Var b) const { return getEqNode(bodies_, b);}
+	void          updateRule(PrgRule& r);
 	template <class C>
 	Var getEqNode(C& vec, Var id)  const {
 		if (!vec[id]->eq()) return id;
@@ -425,7 +517,6 @@ private:
 	void          setConflict();
 	void          updateFrozenAtoms(const Solver&);
 	void          normalize();
-	uint32        startAtom() const { return incData_?incData_->startAtom_:1; }
 	// ------------------------------------------------------------------------
 	// Statistics
 	void upRules(RuleType r, int i)   { stats.rules[0] += i; stats.rules[r] += i; }
@@ -457,10 +548,10 @@ private:
 	} *minimize_;               // list of minimize-rules
 	struct Incremental  {
 		Incremental();
-		bool frozen(Var atom) const;
 		uint32  startAtom_;       // first atom of current iteration
 		uint32  startVar_;        // first var of the current iteration
 		uint32  startAux_;        // first aux atom of current iteration
+		uint32  startScc_;        // first scc of current iteration
 		VarVec  freeze_;          // list of frozen atoms
 		VarVec  unfreeze_;        // list of atom that are unfreezed in this iteration
 	}* incData_;                // additional state to handle incrementally defined programs 
@@ -486,14 +577,14 @@ private:
  */
 class PrgNode {
 public:
-	static const uint32 noScc = (1U << 29) - 1;
+	static const uint32 noScc = (1U << 28) - 1;
 	
 	//! creates a new node
 	/*!
 	 * \note The ctor creates a node that corresponds to a literal that is false
 	 */
 	explicit PrgNode(Literal lit = negLit(sentVar)) 
-		: lit_(lit), root_(0), value_(value_free), scc_(noScc), eq_(0), seen_(0), ignore_(0) {}
+		: lit_(lit), root_(0), value_(value_free), scc_(noScc), eq_(0), seen_(0), ignore_(0), frozen_(0) {}
 	
 	//! returns true if node has an associated variable in a solver
 	/*!
@@ -540,6 +631,9 @@ public:
 	 */
 	bool      ignore()    const { return ignore_ != 0; }
 
+	//! returns true if this atom is frozen, i.e. to be defined in a later step
+	bool      frozen()    const { return frozen_ != 0; }
+
 	//! returns the id of the node to which this node is equivalent
 	/*! 
 	 * \pre eq() is true
@@ -555,7 +649,7 @@ public:
 	void    setLiteral(Literal x)   { lit_ = x; }
 	void    clearLiteral(bool clVal){ lit_ = negLit(sentVar); if (clVal) value_ = value_free; }
 	bool    setValue(ValueRep v)    {
-		if (v == value_weak_true && ignore()) { v = value_true; }
+		if (v == value_weak_true && ignore() && !frozen()) { v = value_true; }
 		if (value_ == value_free || v == value_ || (value_ == value_weak_true && v == value_true)) {
 			value_ = v;
 			return true;
@@ -574,6 +668,7 @@ public:
 		return true;
 	}
 	void    setIgnore(bool b)       { ignore_ = (uint32)b; }
+	void    setFrozen(bool b)       { frozen_ = (uint32)b; }
 	void    setEq(uint32 oId)       {
 		root_   = oId;
 		eq_     = 1;
@@ -604,6 +699,17 @@ public:
 		}
 	}
 	//@}
+protected:
+	void assign(const PrgNode* other) {
+		lit_   = other->lit_;
+		root_  = other->root_;
+		value_ = other->value_;
+		scc_   = other->scc_;
+		eq_    = other->eq_;
+		seen_  = other->seen_;
+		ignore_= other->ignore_;
+		frozen_= other->frozen_;
+	}
 private:
 	PrgNode(const PrgNode&);
 	PrgNode& operator=(const PrgNode&);
@@ -611,16 +717,16 @@ private:
 	uint32  root_   : 30; // used twofold:
 	                      //  - relevant nodes (i.e. eq() == false):
 	                      //     depth search index used to find root of a SCC and
-												//     afterwards index of this node in the unfounded set checker (if relevant)
+	                      //     afterwards index of this node in the unfounded set checker (if relevant)
 	                      //  - eq nodes (i.e. eq() == true):
 	                      //     id of node to which this node is equivalent
 	uint32  value_  :  2; // truth-value assigned to the node (either compute or derived during preprocessing)
-	uint32  scc_    : 29; // component id of this node (noScc if trivially connected)
+	uint32  scc_    : 28; // component id of this node (noScc if trivially connected)
 	uint32  eq_     :  1; // node is eq to some other node (id is stored in root_)
 	uint32  seen_   :  1; // SCC-check: true if node was already visited
 	uint32  ignore_ :  1; // ignore during scc check (node is false or has a support that is true)
 	                      // For bodies: body is no longer relevant (e.g. one of its subgoals is known to be false)
-	
+	uint32  frozen_ :  1; // For atoms: frozen, i.e. defined in a later incremental step
 };
 
 //! A body-node represents a rule-body in a body-atom-dependency graph.
@@ -853,6 +959,14 @@ public:
 	typedef std::pair<bool, uint32> SimpRes;
 	SimpRes simplifyBodies(Var atomId, ProgramBuilder& prg, bool strong);
 	bool    hasPred(Var bodyId) const { return std::find(preds.begin(), preds.end(), bodyId) != preds.end(); }
+
+	void    assign(const PrgAtomNode* a) {
+		PrgNode::assign(a);
+		posDep = a->posDep;
+		negDep = a->negDep;
+		preds  = a->preds;
+	}
+
 	VarVec    posDep;     // Bodies in which this atom occurs positively
 	VarVec    negDep;     // Bodies in which this atom occurs negatively
 	VarVec    preds;      // Bodies having this atom as head
