@@ -324,7 +324,7 @@ public:
 	 *   \code
 	 *   bool MyPostProp::propagate(Solver& s) {
 	 *     while (!F.empty()) {
-	 *       LitVec C = f.pop();
+	 *       LitVec C = F.pop();
 	 *       if (!integrateClause(C, addSat)) return false;
 	 *     }
 	 *     return true;
@@ -334,7 +334,7 @@ public:
 	 *   \code
 	 *   bool MyPostProp::propagateFixpoint(Solver& s) {
 	 *     while (!F.empty()) {
-	 *       LitVec C = f.pop();
+	 *       LitVec C = F.pop();
 	 *       if (!integrateClause(C, addSat)) return false;
 	 *       if (s.queueSize() > 0 && !s.propagateUntil(this)) return false;
 	 *     }
@@ -468,6 +468,11 @@ public:
 		assert(validVar(v));
 		return assign_.value(v);
 	}
+	//! returns the final value of v or value_free if v is not yet unconditionally assigned 
+	ValueRep finalValue(Var v) const {
+		return value(v) == value_free || level(v) > 0 ? value_free : value(v);
+	}
+
 	//! returns the previous value of v
 	ValueRep savedValue(Var v) const {
 		assert(validVar(v));
@@ -557,9 +562,9 @@ public:
 	const LitVec& assignment() const { return assign_.trail; }
 	
 	//! returns true, if the current assignment is conflicting
-	bool hasConflict() const { return !conflict_.empty(); }
+	bool hasConflict() const { return !conflict_.lits.empty(); }
 	//! returns the current conflict as a set of literals
-	const LitVec& conflict() const { return conflict_; }
+	const LitVec& conflict() const { return conflict_.lits; }
 	//! returns the most recently derived conflict clause
 	const LitVec& conflictClause() const { return cc_; }
 	
@@ -652,9 +657,12 @@ public:
 	//! Marks the literals in cfl as conflicting.
 	/*! 
 	 * \pre !cfl.empty() && !hasConflict()
+	 * \pre computeLevel or cfl has at least one literal from current DL
+	 * \param cfl The set of (true) literals that is conflicting
+	 * \param computeLevel compute max decision level of cfl
 	 * \post cfl.empty() && hasConflict()
 	 */
-	void setConflict(LitVec& cfl) { conflict_.swap(cfl); }
+	void setConflict(LitVec& cfl, bool computeLevel = true);
 
 	//! Sets a conflict that forces the solver to terminate its search
 	/*!
@@ -730,7 +738,7 @@ public:
 	 * If the SearchStrategy is set to learning, resolveConflict implements
 	 * First-UIP learning and backjumping. Otherwise it simply applies
 	 * chronological backtracking.
-	 * \pre hasConflict
+	 * \pre hasConflict()
 	 * \return
 	 *  - true if the conflict was successfully resolved
 	 *  - false otherwise
@@ -739,6 +747,12 @@ public:
 	 */
 	bool resolveConflict();
 
+	//! analyzes the active root-level conflict and stores the result in conflictClause()
+	/*!
+	 * \pre hasConflict() && rootLevel() == decisionLevel()
+	 */
+	void analyzeRootConflict(Literal a = posLit(0));
+
 	//! backtracks the last decision and sets the backtrack-level to the resulting decision level.
 	/*!
 	 * \return
@@ -832,6 +846,9 @@ private:
 		PostPropagator* look;
 		PostPropagator* saved;
 	};
+	void    extractReason(Literal p, LitVec& out) {
+		stats.solve.updateAntes(reason(p).reason(p, out));
+	}
 	inline  bool hasStopConflict() const;
 	bool    validWatch(Literal p) const { return p.index() < (uint32)watches_.size(); }
 	uint32  mark(uint32 s, uint32 e);
@@ -853,11 +870,19 @@ private:
 	ConstraintDB      constraints_; // problem constraints
 	ConstraintDB      learnts_;     // learnt constraints
 	ImpliedLits       impliedLits_; // Lits that were asserted on current dl but are logically implied earlier
-	LitVec            conflict_;    // stores conflict-literals for later analysis
 	LitVec            cc_;          // temporary: stores conflict clause within analyzeConflict
 	TrailLevels       levels_;      // Stores start-position in trail of every decision level
 	Watches           watches_;     // watches_[p.index()] is a list of constraints watching p
 	PPList            post_;        // (possibly empty) list of post propagators
+	struct Conflict {
+		Conflict() : level(UINT32_MAX) {}
+		void clear() { 
+			level = UINT32_MAX;
+			lits.clear();
+		}
+		LitVec          lits;       // conflict-literals for analysis
+		uint32          level;      // current conflict level (highest dl in lits)
+	}                 conflict_;    // stores current conflict for later analysis
 	DecisionHeuristic*randHeuristic_;
 	VarVec*           levConflicts_;// For a DL d, levConflicts_[d-1] stores num conflicts when d was started
 	ConstraintDB*     undoHead_;    // Free list of undo DBs
