@@ -44,7 +44,7 @@ ClauseCreator::ClauseCreator(Solver* s)
 : solver_(s) { }
 
 void ClauseCreator::init(ConstraintType t) {
-	assert(!solver_ || solver_->decisionLevel() == 0 || t != Constraint_t::native_constraint);
+	assert(!solver_ || solver_->decisionLevel() == 0 || t != Constraint_t::static_constraint);
 	literals_.clear();
 	fwLev_     = 0;
 	swLev_     = 0;
@@ -155,16 +155,16 @@ bool ClauseCreator::end(bool addSat, Constraint** out) {
 }
 
 bool ClauseCreator::addClause(Solver& s, ConstraintType type, const LitVec& lits, uint32 sw, Constraint** out) {
-	assert(s.decisionLevel() == 0 || type != Constraint_t::native_constraint);
+	assert(s.decisionLevel() == 0 || type != Constraint_t::static_constraint);
 	if (out) *out = 0;
 	if (lits.empty()) return s.addUnary(negLit(0)); // UNSAT on level 0
-	if (type == Constraint_t::native_constraint && s.strategies().satPrePro.get()) {
+	if (type == Constraint_t::static_constraint && s.strategies().satPrePro.get()) {
 		return s.strategies().satPrePro->addClause(lits);
 	}
 	bool unit = lits.size() == 1 || s.isFalse(lits[sw]);
 	s.strategies().heuristic->newConstraint(s, &lits[0], lits.size(), type);
 	if (lits.size() < 4) {
-		if (type != Constraint_t::native_constraint) s.stats.solve.addLearnt((uint32)lits.size(), type);
+		if (type != Constraint_t::static_constraint) s.stats.solve.addLearnt((uint32)lits.size(), type);
 		if (lits.size() == 1)      { return s.addUnary(lits[0]);    }
 		else if (lits.size() == 2) { s.addBinary(lits[0], lits[1]); }
 		else                       { s.addTernary(lits[0], lits[1], lits[2]); }
@@ -178,7 +178,7 @@ bool ClauseCreator::addClause(Solver& s, ConstraintType type, const LitVec& lits
 		Constraint* newCon;
 		Literal first  = lits[0];
 		Literal second = lits[sw];
-		if (type == Constraint_t::native_constraint) {
+		if (type == Constraint_t::static_constraint) {
 			newCon = Clause::newClause(s, lits);
 			s.add(newCon);
 		}
@@ -199,11 +199,11 @@ bool ClauseCreator::addClause(Solver& s, ConstraintType type, const LitVec& lits
 Constraint* Clause::newClause(Solver& s, const LitVec& lits) {
 	LitVec::size_type nl = 2 + lits.size(); // 2 sentinels
 	void* mem = ::operator new( sizeof(Clause) + (nl*sizeof(Literal)) );
-	return new (mem) Clause(s, lits, 0, Constraint_t::native_constraint, 0);
+	return new (mem) Clause(s, lits, 0, Constraint_t::static_constraint, 0);
 }
 
 LearntConstraint* Clause::newLearntClause(Solver& s, const LitVec& lits, ConstraintType t, LitVec::size_type secondWatch) {
-	assert(t != Constraint_t::native_constraint);
+	assert(t != Constraint_t::static_constraint);
 	LitVec::size_type nl = 3 + lits.size(); // 2 sentinels + 1 lit for activity
 	void* mem = ::operator new( sizeof(Clause) + (nl * sizeof(Literal)) );
 	return new (mem) Clause(s, lits, secondWatch, t, 0);
@@ -223,7 +223,7 @@ Clause::Clause(Solver& s, const LitVec& theLits, LitVec::size_type sw, Constrain
 	end()->asUint()   = 1;  // Ending Sentinel    - var 0 + watch-flag on
 	other_  = theLits[0];
 	std::memcpy(begin(), &theLits[0], sizeof(Literal) * theLits.size());
-	if (t == Constraint_t::native_constraint) {
+	if (t == Constraint_t::static_constraint) {
 		initWatches(s);
 	}
 	else {
@@ -354,7 +354,7 @@ Constraint::PropResult Clause::propagate(const Literal&, uint32& data, Solver& s
 	}
 }
 
-void Clause::reason(const Literal& p, LitVec& rLits) {
+ConstraintType Clause::reason(const Literal& p, LitVec& rLits) {
 	assert(other_ == p);
 	const Literal* e = end();
 	for (Literal* r = begin(); r != e; ++r) {
@@ -369,6 +369,7 @@ void Clause::reason(const Literal& p, LitVec& rLits) {
 		}
 	}
 	bumpActivity();
+	return Clause::type();
 }
 
 bool Clause::simplify(Solver& s, bool reinit) {
@@ -610,7 +611,7 @@ Constraint::PropResult LoopFormula::propagate(const Literal&, uint32& data, Solv
 
 // Body: all other bodies + active atom
 // Atom: all bodies
-void LoopFormula::reason(const Literal& p, LitVec& lits) {
+ConstraintType LoopFormula::reason(const Literal& p, LitVec& lits) {
 	// all relevant bodies
 	for (uint32 x = 1; x != end_-1; ++x) {
 		if (lits_[x] != p) {
@@ -622,6 +623,7 @@ void LoopFormula::reason(const Literal& p, LitVec& lits) {
 		lits.push_back(~lits_[end_-1]);
 	}
 	++activity_;
+	return Constraint_t::learnt_loop;
 }
 
 uint32 LoopFormula::size() const {
