@@ -39,62 +39,63 @@ Literal Preprocessor::newBodyLit(PrgBodyNode* n) {
 	else                  { return posLit(false_); }
 }
 
-// If the program is defined incrementally, marks atoms from previous steps as supported
+// If the program is defined incrementally, marks atoms from previous steps as supported.
+// Logically, the functions acts as if the program contains a choice rule over
+// all atoms defined in the previous steps minus those to be unfrozen and hence
+// defined in this step.	
 // Pre: Atoms that are true/false have their value-member set
 void Preprocessor::updatePreviouslyDefinedAtoms(Var startAtom, bool strong) {
-	for (Var i = 0; i != startAtom; ++i) {
-		PrgAtomNode* a = prg_->atoms_[i];
-		if (!strong && !a->visited()) {
-			// a is an atom about to be unfrozen
-			a->preds.clear();
+	follow_.clear();
+	if (prg_->incData_) {
+		// mark all atom to be unfrozen in this step
+		for (VarVec::const_iterator it = prg_->incData_->unfreeze_.begin(), end = prg_->incData_->unfreeze_.end(); it != end; ++it) {
+			assert(*it < startAtom);
+			PrgAtomNode* a = prg_->atoms_[*it];
+			Literal      x = a->literal(); x.watch();
+			a->setLiteral(x);
 		}
-		PrgAtomNode* aEq = 0;
-		if (strong) {
-			nodes_[i].aSeen = 1;
-			uint32 eq = prg_->getEqAtom(i);
-			if (eq != i) {
-				aEq = prg_->atoms_[eq];
-				a->clearLiteral(true);
-				a->setValue( aEq->value() );
-			}
+	}
+	LitVec::size_type old = prg_->compute_.size();
+	for (Var i = 1; i < startAtom; ++i) {
+		PrgAtomNode* a   = prg_->atoms_[i];
+		if (a->literal().watched()) { 
+			Literal x = a->literal(); x.clearWatch(); 
+			a->setLiteral(x); 
+			if (!strong) { a->preds.clear(); }
+			else         { setRootAtom(a->literal(), i); }
 		}
-		ValueRep v = a->hasVar() || aEq ? a->value() : value_false;
-		for (VarVec::size_type b = 0; b != a->posDep.size(); ++b) {
-			PrgBodyNode* body = prg_->bodies_[a->posDep[b]];
-			if (strong && (aEq || v != value_free)) {
-				nodes_[a->posDep[b]].sBody = 1;
+		else if (!a->eq() && a->value() != value_false) {
+			if (strong) {
+				setRootAtom(a->literal(), i);
+				nodes_[i].aSeen = 1;
+				propagateAtomVar((Var)i, a);
 			}
-			if (v != value_false) {
-				if (strong) ++nodes_[a->posDep[b]].known;
-				bool isSupp = body->isSupported();
-				body->onPosPredSupported(Var(i));
-				if (!isSupp && body->isSupported()) {
-					prg_->initialSupp_.push_back(a->posDep[b]);
+			else {
+				for (VarVec::size_type b = 0; b != a->posDep.size(); ++b) {
+					PrgBodyNode* body = prg_->bodies_[a->posDep[b]];
+					bool isSupp = body->isSupported();
+					body->onPosPredSupported(Var(i));
+					if (!isSupp && body->isSupported()) {
+						prg_->initialSupp_.push_back(a->posDep[b]);
+					}
 				}
 			}
 		}
-		if (strong) {
-			for (VarVec::size_type b = 0; b != a->negDep.size(); ++b) {
-				if (aEq || v != value_free) {
-					nodes_[a->negDep[b]].sBody = 1;
-				}
-				if (v != value_true) {
-					++nodes_[a->negDep[b]].known;
-				}
-			}
-			if (v != value_free) {
-				a->posDep.clear();
-				a->negDep.clear();
+		else {
+			assert(!a->eq() || (a->posDep.empty() && a->negDep.empty() && "EQ-Atom in new Program - rule not simplified!"));
+			if (a->posDep.size() + a->negDep.size() + a->preds.size() > 0) {
+				prg_->setCompute(i, false);
 			}
 		}
 	}
+	applyCompute(prg_->compute_, old, strong);
 }
 
 // Back-propagates atoms from the compute statement.
 bool Preprocessor::applyCompute(LitVec& compute, uint32 start, bool strong) {
 	uint32 newStart = (uint32)compute.size();
 	for (LitVec::size_type i = start; i < compute.size(); ++i) {
-		Var        atomId = compute[i].var();
+		Var        atomId = prg_->getEqAtom(compute[i].var());
 		PrgAtomNode* atom = prg_->atoms_[atomId];
 		if (strong) {
 			// We can safely remove atom from bodies where it occurs negatively.
@@ -132,8 +133,7 @@ bool Preprocessor::applyCompute(LitVec& compute, uint32 start, bool strong) {
 			atom->preds.clear();
 		}
 		else if (!atom->setValue(value_weak_true)) return false;
-		else if (backprop_ && atom->preds.size() < 2) {
-			if (atom->preds.empty()) return false;
+		else if (backprop_ && atom->preds.size() == 1) {
 			PrgBodyNode* b = prg_->bodies_[atom->preds[0]];
 			if (!b->setValue(value_weak_true) || !b->backpropagate(*prg_, compute)) return false;
 		}
@@ -164,18 +164,19 @@ bool Preprocessor::preprocessSimple(bool bodyEq) {
 		PrgBodyNode* b = prg_->bodies_[prg_->initialSupp_[i]];
 		b->buildHeadSet();
 		if (!bodyEq) b->setLiteral(newBodyLit(b));
+		bool m = bodyEq && b->size() == 0 && !b->isChoice();
 		for (VarVec::const_iterator h = b->heads.begin(); h != b->heads.end(); ++h) {
 			if ( !prg_->atoms_[*h]->hasVar() ) {
 				prg_->atoms_[*h]->preds.clear();
-				bool m = bodyEq && b->size() == 0 && !b->isChoice();
 				if (m) {
 					prg_->atoms_[*h]->setLiteral(posLit(0));
 					prg_->atoms_[*h]->setValue(value_true);
-					m = true;
 				}
 				else {
 					prg_->atoms_[*h]->setLiteral(posLit(prg_->vars_.add(Var_t::atom_var)));
 				}
+			}
+			if (prg_->atoms_[*h]->preds.empty()) {
 				const VarVec& bl = prg_->atoms_[*h]->posDep;
 				for (VarVec::const_iterator it = bl.begin(); it != bl.end(); ++it) {
 					PrgBodyNode* nb = prg_->bodies_[*it];
@@ -232,17 +233,17 @@ bool Preprocessor::preprocessEq(uint32 maxIters) {
 	Var startAtom   = prg_->incData_?prg_->incData_->startAtom_ : 0;
 	Var stopAtom    = startAtom;
 	Var startVar    = prg_->incData_?prg_->incData_->startVar_  : 1;
-	pass_     = 0;
-	maxPass_  = maxIters;
+	pass_           = 0;
+	maxPass_        = maxIters;
 	if (!applyCompute(prg_->compute_, 0, true)) return false;
 	do {
 		++pass_;
-		updatePreviouslyDefinedAtoms(startAtom, true);
-		std::stable_sort(prg_->initialSupp_.begin(), prg_->initialSupp_.end(), LessBodySize(prg_->bodies_));
 		prg_->vars_.shrink(startVar);
 		if (prg_->vars_.size() <= false_) false_ = 0;
 		litToNode_.clear();
-		for (Var i = startAtom; i < stopAtom; ++i) { prg_->atoms_[i]->clearLiteral(false); } 
+		for (Var i = startAtom; i < stopAtom; ++i) { 
+			prg_->atoms_[i]->clearLiteral(false); 
+		} 
 		if (!classifyProgram(startAtom, stopAtom)) return false;
 	} while (stopAtom != (uint32)prg_->atoms_.size());
 	// replace equivalent atoms in minimize rules
@@ -250,7 +251,13 @@ bool Preprocessor::preprocessEq(uint32 maxIters) {
 		for (WeightLitVec::iterator it = r->lits_.begin(); it != r->lits_.end(); ++it) {
 			it->first = Literal(prg_->getEqAtom(it->first.var()), it->first.sign());
 		}
-	}	
+	}
+	// replace equivalent atoms in compute statement
+	for (LitVec::iterator it = prg_->compute_.begin(), end = prg_->compute_.end(); it != end; ++it) {
+		if (prg_->getEqAtom(it->var()) != it->var()) {
+			*it = Literal(prg_->getEqAtom(it->var()), it->sign());
+		}
+	}
 	return true;
 }
 
@@ -258,19 +265,13 @@ bool Preprocessor::preprocessEq(uint32 maxIters) {
 // of a program.
 bool Preprocessor::classifyProgram(uint32 startAtom, uint32& stopAtom) {
 	assert(ok_);
-	follow_.clear();
 	Var bodyId, bodyEqId; PrgBodyNode* body;
 	Var atomId, atomEqId; PrgAtomNode* atom;
-	// classify only supported bodies
-	for (VarVec::size_type i = 0; i < prg_->initialSupp_.size(); ++i) {
-		bodyId  = prg_->initialSupp_[i];
-		body    = prg_->bodies_[bodyId];
-		if (body->visited() || body->ignore()) {
-			if (body->ignore() && body->hasVar()) { body->clearLiteral(false); }
-			continue;
-		}
-		follow_.push_back(bodyId);
-		VarVec::size_type index = 0;
+	VarVec::size_type index = 0;
+	follow_.clear();
+	updatePreviouslyDefinedAtoms(startAtom, true);
+	std::stable_sort(prg_->initialSupp_.begin(), prg_->initialSupp_.end(), LessBodySize(prg_->bodies_));
+	for (VarVec::size_type i = 0;;) {
 		while ( (bodyId = nextBodyId(index)) != varMax ) {
 			body        = prg_->bodies_[bodyId];
 			bodyEqId    = addBodyVar(bodyId, body);
@@ -285,7 +286,7 @@ bool Preprocessor::classifyProgram(uint32 startAtom, uint32& stopAtom) {
 					// remove atom from head of body - equivalent atoms don't need definitions
 					setSimplifyHeads(bodyEqId);
 				}
-				if (bodyId != bodyEqId) {
+				if (bodyId != bodyEqId && atomId == atomEqId) {
 					// body is replaced with bodyEqId - move heads to bodyEqId
 					prg_->bodies_[bodyEqId]->heads.push_back( atomEqId );
 					// mark atom so that the link between atomEqId and bodyId can be replaced
@@ -303,6 +304,20 @@ bool Preprocessor::classifyProgram(uint32 startAtom, uint32& stopAtom) {
 			}
 		}
 		follow_.clear();
+		index = 0;
+		// select next unclassified supported body
+		for (; i < prg_->initialSupp_.size(); ++i) {
+			bodyId  = prg_->initialSupp_[i];
+			body    = prg_->bodies_[bodyId];
+			if (!body->visited() && !body->ignore()) {
+				follow_.push_back(bodyId);
+				break;
+			}
+			else if (body->ignore() && body->hasVar()) {
+				body->clearLiteral(false);
+			}
+		}
+		if (follow_.empty()) break;
 	}
 	return simplifyClassifiedProgram(startAtom, stopAtom);
 }
@@ -374,35 +389,30 @@ bool Preprocessor::simplifyClassifiedProgram(uint32 startAtom, uint32& stopAtom)
 			}
 		}
 	}
-	for (uint32 i = startAtom; i != (uint32)prg_->atoms_.size(); ++i) {
-		PrgAtomNode* a = prg_->atoms_[i];
-		if (a->hasVar()) {
-			PrgAtomNode::SimpRes r(true, UINT32_MAX);
-			if (nodes_[i].asBody == 1 && (r = a->simplifyBodies(Var(i), *prg_, true)).first == false) {
-				return false;
-			}
-			nodes_[i].asBody = 0; nodes_[i].aSeen = 0;
-			if (backprop_ && a->value() != value_free && pass_ != maxPass_) {
-				if (a->value() == value_false && a->preds.empty()) {
-					prg_->compute_.push_back(negLit(Var(i)));
-				}
-				else if (a->value() == value_weak_true && a->preds.size() == 1) {
-					PrgBodyNode* b = prg_->bodies_[a->preds[0]];
-					if (b->value() != value_weak_true) {
-						if (!b->setValue(value_weak_true) || !b->backpropagate(*prg_, prg_->compute_)) { return false; }
-					}
-				}
-				if (prg_->compute_.size()>compute) {
-					stopAtom = std::min(stopAtom,i);
-					a->clearLiteral(false);
-				}
-			}
-			if (stopAtom != prg_->atoms_.size() || (pass_ != maxPass_ && reclassify(a, i, r.second))) {
-				stopAtom = std::min(stopAtom,i);
-				a->clearLiteral(false);
+	if (VarVec* unfreeze = prg_->incData_ ? &prg_->incData_->unfreeze_ : 0) {
+		for (VarVec::const_iterator it = unfreeze->begin(), end = unfreeze->end(); it != end; ++it) {
+			ValueRep x = simplifyAtom(*it, false);
+			if (x != value_true) {
+				if (x == value_false)  { return false; }
+				if (pass_ != maxPass_) { stopAtom = 0; }
+				stopAtom = 0;
 			}
 		}
 	}
+	bool more = (pass_ != maxPass_ && stopAtom == prg_->atoms_.size());
+	bool clear= stopAtom != prg_->atoms_.size();
+	for (uint32 i = startAtom; i != (uint32)prg_->atoms_.size(); ++i) {
+		ValueRep x = simplifyAtom(Var(i), more);
+		if (x != value_true) {
+			if (x == value_false) { return false; }
+			if (more) {
+				more     = false;
+				clear    = true;
+				stopAtom = i;
+			}
+		}
+		if (clear) { prg_->atoms_[i]->clearLiteral(false); }
+	}
 	return applyCompute(prg_->compute_,(uint32)compute,true);
 }
 
@@ -444,6 +454,39 @@ void Preprocessor::newFalseBody(PrgBodyNode* body, uint32 oldHash) {
 	removeFromIndex(body, oldHash);
 }
 
+// Simplify the classified atom with the given id.
+// Update list of bodies defining this atom and check
+// if atom has a distinct var although it is eq to some body.
+// Return:
+//  value_false    : conflict
+//  value_true     : ok
+//  value_weak_true: ok but atom should be reclassified
+ValueRep Preprocessor::simplifyAtom(Var id, bool reclass) {
+	if (PrgAtomNode* a = prg_->atoms_[id]->hasVar() ? prg_->atoms_[id] : 0) {
+		ValueRep v   = a->value();
+		ValueRep ret = value_true;
+		PrgAtomNode::SimpRes r(true, UINT32_MAX);
+		if (nodes_[id].asBody == 1 && (r = a->simplifyBodies(id, *prg_, true)).first == false) {
+			return value_false;
+		}
+		nodes_[id].asBody = 0; nodes_[id].aSeen = 0;
+		if (a->value() != v) {
+			if (a->value() == value_false) {
+				prg_->compute_.push_back(negLit(id));
+				ret = value_weak_true;
+			}
+			else if (a->value() == value_weak_true) {
+				prg_->compute_.push_back(posLit(id));
+			}
+		}
+		if (reclass && reclassify(a, id, r.second)) {
+			ret = value_weak_true;
+		}
+		return ok_ ? ret : value_false;
+	}
+	return value_true;
+}
+
 // check if atom has a distinct var although it is eq to some body
 bool Preprocessor::reclassify(PrgAtomNode* a, uint32 atomId, uint32 diffLits) {
 	if ((a->preds.empty() && a->hasVar()) || 
@@ -459,7 +502,7 @@ bool Preprocessor::reclassify(PrgAtomNode* a, uint32 atomId, uint32 diffLits) {
 		bool stableTruth= eq->value() == value_true || eq->value() == value_false;
 		a->setLiteral(x);
 		if (!prg_->mergeEqAtoms(atomId, getRootAtom(x))) {
-			return false;
+			return ok_ = false;
 		}
 		// update bodies containing a	
 		LitVec temp; temp.reserve(a->posDep.size() + a->negDep.size());
@@ -477,15 +520,19 @@ bool Preprocessor::reclassify(PrgAtomNode* a, uint32 atomId, uint32 diffLits) {
 				bool wasSup = temp[i].sign() || bn->isSupported();
 				std::pair<uint32, uint32> hash;
 				if (!bn->simplifyBody(*prg_, bodyId, hash, *this, true)) {
-					return false;
+					return ok_ = false;
 				}
 				if (!bn->simplifyHeads(*prg_, *this, true)) {
-					return false;
+					return ok_ = false;
 				}
 				removeFromIndex(bn, hash.first);
 				uint32 otherId = findEqBody(bn, hash.second);
 				if (otherId != varMax) {
 					mergeBodies(bn, otherId);
+					if (nodes_[otherId].sHead == 1 && !prg_->bodies_[otherId]->simplifyHeads(*prg_, *this, true)) {
+						return ok_ = false;
+					}
+					nodes_[otherId].sHead = 0;
 				}
 				else {
 					prg_->bodyIndex_.insert(ProgramBuilder::BodyIndex::value_type(hash.second, bodyId));
@@ -588,31 +635,6 @@ uint32 Preprocessor::addBodyVar(Var bodyId, PrgBodyNode* body) {
 // associates a variable with the atom if necessary
 // b is the supported body that brings a into the closure
 uint32 Preprocessor::addAtomVar(Var atomId, PrgAtomNode* a, PrgBodyNode* b) {
-	if ( nodes_[atomId].aSeen == 1 ) { 
-		if (hasBody(b->literal()) && a->preds.size() > 1) {
-			// The eq-class of b already contains a body. Check 
-			// if it is one of the other bodies defining a.
-			nodes_[atomId].asBody = 1;
-		}
-		return prg_->getEqAtom(atomId);   
-	}
-	nodes_[atomId].aSeen = 1;
-	uint32 retId = atomId;
-	// if set of defining bodies of this atom has changed, update it...
-	if (nodes_[atomId].asBody == 1 && !a->simplifyBodies(atomId, *prg_, false).first) {
-		ok_ = false;
-		return atomId;
-	}
-	if (pass_ == 1) { nodes_[atomId].asBody = 1; }
-	if (!b->isChoice() && (b->value() == value_true || a->preds.size() == 1)) {
-		// Atom is equivalent to b
-		prg_->incEqs(Var_t::atom_body_var);
-		a->setLiteral( b->literal() );
-		prg_->vars_.setAtomBody(a->var());
-	}
-	else {
-		a->setLiteral( posLit(prg_->vars_.add(Var_t::atom_var)) ); 
-	}
 	if (b->value() == value_true) {
 		// Since b is true, it is always a valid support for a, a can never become unfounded. 
 		// So ignore it during SCC check and unfounded set computation.
@@ -622,28 +644,56 @@ uint32 Preprocessor::addAtomVar(Var atomId, PrgAtomNode* a, PrgBodyNode* b) {
 			return atomId;
 		}
 	}
-	PrgAtomNode* eqAtom = 0;
-	Var eqAtomId = getRootAtom(a->literal());
-	if ( eqAtomId == varMax ) {
-		setRootAtom(a->literal(), atomId);
-	}
-	else {
-		eqAtom    = prg_->atoms_[eqAtomId];
-		if (!prg_->mergeEqAtoms(atomId, eqAtomId)) {
+	if (nodes_[atomId].aSeen == 0) { 
+		nodes_[atomId].aSeen = 1;
+		// if set of defining bodies of this atom has changed, update it...
+		if (nodes_[atomId].asBody == 1 && !a->simplifyBodies(atomId, *prg_, false).first) {
 			ok_ = false;
 			return atomId;
 		}
-		retId = eqAtomId;
+		if (pass_ == 1) { nodes_[atomId].asBody = 1; }
+		if (!a->hasVar()) {
+			if (!b->isChoice() && (b->value() == value_true || a->preds.size() == 1)) {
+				// Atom is equivalent to b
+				prg_->incEqs(Var_t::atom_body_var);
+				a->setLiteral( b->literal() );
+				prg_->vars_.setAtomBody(a->var());
+			}
+			else {
+				a->setLiteral( posLit(prg_->vars_.add(Var_t::atom_var)) ); 
+			}
+			Var eqAtomId = getRootAtom(a->literal());
+			if ( eqAtomId == varMax ) {
+				setRootAtom(a->literal(), atomId);
+			}	
+			else if (!prg_->mergeEqAtoms(atomId, eqAtomId)) {
+				ok_ = false;
+				return atomId;
+			}
+			assert(getRootAtom(a->literal()) != varMax);
+		}
+		return propagateAtomVar(atomId, a);
 	}
+	else if (hasBody(b->literal()) && a->preds.size() > 1) {
+		// The eq-class of b already contains a body. Check 
+		// if it is one of the other bodies defining a.
+		nodes_[atomId].asBody = 1;
+	}
+	return prg_->getEqAtom(atomId);   
+}
+
+uint32 Preprocessor::propagateAtomVar(Var atomId, PrgAtomNode* a) {
 	// If atom a has a truth-value or is eq to a', we'll remove
 	// it from all bodies. If there is an atom x, s.th. a.lit == ~x.lit, we mark all
 	// bodies containing both a and x for simplification in order to detect
 	// duplicates/contradictory body-literals.
 	// In case that a == a', we also mark all bodies containing a
 	// for head simplification in order to detect rules like: a' :- a,B. and a' :- B,not a.
+	uint32 eqId       = prg_->getEqAtom(atomId);
+	PrgAtomNode* eqAt = prg_->atoms_[eqId];
 	bool stableTruth  = a->value() == value_true || a->value() == value_false;
-	bool removeAtom   = stableTruth || eqAtom != 0;
-	bool removeNeg    = removeAtom || a->value() == value_weak_true;
+	bool removeAtom   = stableTruth || eqAt != a;
+	bool removeNeg    = removeAtom  || a->value() == value_weak_true;
 	if (!removeAtom && getRootAtom(~a->literal()) != varMax) {
 		PrgAtomNode* comp = prg_->atoms_[getRootAtom(~a->literal())];
 		assert(a->literal() == ~comp->literal());
@@ -677,9 +727,9 @@ uint32 Preprocessor::addAtomVar(Var atomId, PrgAtomNode* a, PrgBodyNode* b) {
 		}
 		if (removeAtom || nodes_[bodyId].mBody == 1) { 
 			setSimplifyBody(bodyId);
-			if (eqAtom != 0) {
+			if (eqAt != a) {
 				setSimplifyHeads(bodyId);
-				if (!stableTruth) { eqAtom->posDep.push_back( bodyId ); }
+				if (!stableTruth) { eqAt->posDep.push_back( bodyId ); }
 			}
 		}
 		++i;
@@ -693,9 +743,9 @@ uint32 Preprocessor::addAtomVar(Var atomId, PrgAtomNode* a, PrgBodyNode* b) {
 		}
 		if (removeNeg || nodes_[bodyId].mBody == 1) { 
 			setSimplifyBody(bodyId);
-			if (eqAtom != 0) {
+			if (eqAt != a) {
 				setSimplifyHeads(bodyId);
-				if (!stableTruth && a->value() != value_weak_true) { eqAtom->negDep.push_back( bodyId ); }
+				if (!stableTruth && a->value() != value_weak_true) { eqAt->negDep.push_back( bodyId ); }
 			}
 		}
 		++i;
@@ -712,7 +762,7 @@ uint32 Preprocessor::addAtomVar(Var atomId, PrgAtomNode* a, PrgBodyNode* b) {
 			}
 		}
 	}
-	return retId;
+	return eqId;
 }
 
 // body b has changed - remove old entry from body node index
