@@ -18,7 +18,7 @@
 // Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 //
 
-#ifdef _MSC_VER
+#if defined(_MSC_VER)
 #pragma warning (disable : 4146) // unary minus operator applied to unsigned type, result still unsigned
 #pragma warning (disable : 4996) // 'std::_Fill_n' was declared deprecated
 #endif
@@ -32,6 +32,28 @@
 #include <sstream>
 #include <numeric>
 
+#if defined(_MSC_VER)
+#	if !defined(__FUNCTION__)
+#	define MY_FUNCNAME __FILE__
+#	else
+# define MY_FUNCNAME __FUNCTION__
+# endif
+#elif defined(__GNUC__)
+#define MY_FUNCNAME __PRETTY_FUNCTION__
+#else
+#define MY_FUNCNAME __FILE__
+#endif
+
+std::string precondition_error(const char* ex, const char* func, unsigned line) {
+	std::stringstream err;
+	err << func << "@" << line << ": precondition violated: " << ex;
+	return err.str();
+}
+
+#define check_precondition(x, EX) \
+	(void)( (!!(x)) || (throw EX(precondition_error((#x), MY_FUNCNAME, __LINE__)), 0))
+
+
 
 namespace Clasp {
 /////////////////////////////////////////////////////////////////////////////////////////
@@ -152,7 +174,7 @@ PrgBodyNode::PrgBodyNode(uint32 id, const PrgRule& rule, const PrgRule::RData& r
 			goals_[p] = x;
 			a->posDep.push_back(id);
 			if (w) { 
-				assert(rule.body[i].second>0);
+				check_precondition(rule.body[i].second>0, std::logic_error);
 				spw += (extra_.ext->weights[p] = rule.body[i].second);
 			}
 			++p;
@@ -161,7 +183,7 @@ PrgBodyNode::PrgBodyNode(uint32 id, const PrgRule& rule, const PrgRule::RData& r
 			goals_[posSize_+n]  = x;
 			if (prg.eqIters_ != 0) { a->negDep.push_back(id); }
 			if (w) { 
-				assert(rule.body[i].second>0);
+				check_precondition(rule.body[i].second>0, std::logic_error);
 				snw += (extra_.ext->weights[posSize_+n] = rule.body[i].second);
 			}
 			++n;
@@ -260,7 +282,7 @@ weight_t PrgBodyNode::bound() const {
 weight_t PrgBodyNode::sumWeights() const {
 	return !extended()
 		? (weight_t)size()
-		: (uint32)std::max(extra_.ext->sumWeights, weight_t(0));
+		: (weight_t)std::max(extra_.ext->sumWeights, weight_t(0));
 }
 
 // Returns the weight of the idx'th subgoal in B+/B-
@@ -350,13 +372,10 @@ bool PrgBodyNode::addPredecessorClauses(Solver& s, ClauseCreator& gc, const Atom
 // Remove false/true literals
 bool PrgBodyNode::simplifyBody(ProgramBuilder& prg, uint32 bodyId, std::pair<uint32, uint32>& hashes, Preprocessor& pre, bool strong) {
 	hashes.first = hashes.second = 0;
-	bool ok = sumWeights() >= bound();
+	bool ok  = sumWeights() >= bound();
 	uint32 j = 0;
 	Var eq, comp;
-	bool rem;
-	bool dirty = false;
 	for (uint32 i = 0, end = size_; i != end && ok; ++i) {
-		rem = true;
 		Var      a = goals_[i].var();
 		hashes.first += hashId(goals_[i].sign()?-a:a);
 		if ((eq = prg.getEqAtom(a)) != a) {
@@ -367,16 +386,15 @@ bool PrgBodyNode::simplifyBody(ProgramBuilder& prg, uint32 bodyId, std::pair<uin
 			else {
 				a         = comp;
 				goals_[i] = Literal(comp, true);
-				dirty     = true;
 				prg.atoms_[comp]->negDep.push_back(bodyId);
 				VarVec& eqPos = prg.atoms_[eq]->posDep;
 				VarVec::iterator it = std::find(eqPos.begin(), eqPos.end(), bodyId);
 				if (it != eqPos.end()) eqPos.erase(it);
 			}
 		}
+		bool mark  = false, rem = true;
 		ValueRep v = prg.atoms_[a]->value();
 		Literal p; 
-		bool mark = false;
 		if (prg.atoms_[a]->hasVar() || strong) {
 			p = goals_[i].sign() ? ~prg.atoms_[a]->literal() : prg.atoms_[a]->literal();
 			v = prg.atoms_[a]->hasVar() ? v : value_false;
@@ -437,23 +455,25 @@ bool PrgBodyNode::simplifyBody(ProgramBuilder& prg, uint32 bodyId, std::pair<uin
 			}
 		}
 	}
-	size_       = j;
-	uint32 posS = 0;
-	for (uint32 i = 0, end = (uint32)size_; i != end; ++i) {
-		prg.vars_.unmark( prg.atoms_[goals_[i].var()]->var() );
-		hashes.second += hashId(goals_[i].sign()?-goals_[i].var():goals_[i].var());
-		posS          += goals_[i].sign() == false;
-	}
-	posSize_ = posS;
-	if (dirty) {
-		uint32 j;
-		for (uint32 i = 0; i != posSize_; ++i) {
-			if (goals_[i].sign()) {
-				for (j = i+1; goals_[j].sign(); ++j) { ; }
-				std::swap(goals_[i], goals_[j]);
+	// unmark atoms, compute new hash value,
+	// and restore pos | neg partition in case
+	// we changed some positive goals to negative ones 
+	size_    = j;
+	uint32 p = 0, n = j;
+	for (uint32 a, h; p < n;) {
+		if      (!goals_[p].sign())      { h = hashId( (a = goals_[p++].var()));  }
+		else if (goals_[n-1].sign())     { h = hashId(-(a = goals_[--n].var()));  }
+		else /* restore pos|neg order */ {
+			std::swap(goals_[p], goals_[n-1]);
+			if (hasWeights()) {
+				std::swap(extra_.ext->weights[p], extra_.ext->weights[n-1]);
 			}
+			continue;
 		}
+		prg.vars_.unmark( prg.atoms_[a]->var() );
+		hashes.second += h;
 	}
+	posSize_ = p;
 	assert(sumWeights() >= bound() || !ok);
 	if (!ok) {          // body is false...
 		setIgnore(true);  // ...and therefore can be ignored
@@ -589,17 +609,16 @@ bool PrgBodyNode::backpropagate(ProgramBuilder& prg, LitVec& comp) {
 /////////////////////////////////////////////////////////////////////////////////////////
 class ProgramBuilder::CycleChecker {
 public:
-	CycleChecker(const AtomList& prgAtoms, const BodyList& prgBodies, bool check)
+	CycleChecker(const AtomList& prgAtoms, const BodyList& prgBodies, bool check, uint32 firstScc)
 		: atoms_(prgAtoms)
 		, bodies_(prgBodies)
 		, count_(0)
-		, sccs_(0)
+		, sccs_(firstScc)
 		, check_(check) {
 	}
 	void visit(PrgBodyNode* b) { if (check_) visitDfs(b, true); }
 	void visit(PrgAtomNode* a) { if (check_) visitDfs(a, false); }
-	bool   hasCycles() const     { return sccs_ > 0; }
-	uint32 sccs()      const     { return sccs_; }
+	uint32 sccs()              const { return sccs_; }
 	const AtomList& sccAtoms() const { return sccAtoms_; }
 private:
 	CycleChecker& operator=(const CycleChecker&);
@@ -753,14 +772,15 @@ void PreproStats::accu(const PreproStats& o) {
 /////////////////////////////////////////////////////////////////////////////////////////
 // class ProgramBuilder
 /////////////////////////////////////////////////////////////////////////////////////////
-ProgramBuilder::ProgramBuilder() : minimize_(0), incData_(0), ufs_(0), eqIters_(UINT32_MAX), erMode_(mode_native), eqDfs_(false) { }
-ProgramBuilder::~ProgramBuilder() { disposeProgram(true); }
-ProgramBuilder::Incremental::Incremental() : startAtom_(1), startVar_(1), startAux_(1) {}
-bool ProgramBuilder::Incremental::frozen(Var atom) const {
-	return std::find(freeze_.begin(), freeze_.end(), atom) != freeze_.end();
+ProgramBuilder::ProgramBuilder() 
+	: minimize_(0), incData_(0), ufs_(0), eqIters_(UINT32_MAX), erMode_(mode_native)
+	, eqDfs_(false), normalize_(false), frozen_(true) { 
 }
+ProgramBuilder::~ProgramBuilder() { disposeProgram(true); }
+ProgramBuilder::Incremental::Incremental() : startAtom_(1), startVar_(1), startAux_(1), startScc_(0) {}
 
 void ProgramBuilder::disposeProgram(bool force) {
+	// remove rules
 	std::for_each( bodies_.begin(), bodies_.end(), DestroyObject() );
 	BodyList().swap(bodies_);
 	bodyIndex_.clear();
@@ -786,37 +806,36 @@ void ProgramBuilder::disposeProgram(bool force) {
 		vars_.clear();
 		ruleState_.clear();
 	}
-	stats.reset();
-}
-
-ProgramBuilder& ProgramBuilder::setAtomName(Var atomId, const char* name) {
-	resize(atomId);
-	atomIndex_->addUnique(atomId, Atom(name));
-	return *this;
-}
-
-ProgramBuilder& ProgramBuilder::setCompute(Var atomId, bool pos) {
-	resize(atomId);
-	ValueRep v = pos ? value_weak_true : value_false;
-	compute_.push_back(Literal(atomId, v==value_false));
-	if (!atoms_[atomId]->setValue(v)) {
-		setConflict();
+	else {
+		// clean up atoms
+		// remove rule associations
+		uint32 startAux = incData_ ? incData_->startAux_ : (uint32)atoms_.size();
+		assert(startAux <= atoms_.size());
+		for (VarVec::size_type i = 0; i != startAux; ++i) {
+			if (atoms_[i]->eq() && getEqAtom(i) >= startAux) {
+				// atom i is equivalent to some aux atom 
+				// make i the new root
+				PrgAtomNode* eq = atoms_[getEqAtom(i)];
+				assert(!eq->eq());
+				eq->posDep.clear(); eq->negDep.clear(); eq->preds.clear();
+				atoms_[i]->assign(eq);
+				eq->setEq(i);
+			}
+			VarVec().swap(atoms_[i]->posDep);
+			VarVec().swap(atoms_[i]->negDep);
+			VarVec().swap(atoms_[i]->preds);
+		}
+		// delete any introduced aux atoms
+		// this is safe because aux atoms are never part of the input program
+		// it is necessary in order to free their ids, i.e. the id of an aux atom
+		// from step I might be needed for a program atom in step I+1
+		for (VarVec::size_type i = startAux; i != atoms_.size(); ++i) {
+			delete atoms_[i];
+		}
+		atoms_.erase(atoms_.begin()+startAux, atoms_.end());
 	}
-	return *this;
-}
-
-ProgramBuilder& ProgramBuilder::freeze(Var atomId) {
-	resize(atomId);
-	if (!incData_) incData_ = new Incremental();
-	incData_->freeze_.push_back(atomId);
-	return *this;
-}
-
-ProgramBuilder& ProgramBuilder::unfreeze(Var atomId) {
-	resize(atomId);
-	if (!incData_) incData_ = new Incremental();
-	incData_->unfreeze_.push_back(atomId);
-	return *this;
+	rule_.clear();
+	stats.reset();
 }
 
 ProgramBuilder& ProgramBuilder::startProgram(AtomIndex& index, DefaultUnfoundedCheck* ufs) {
@@ -837,47 +856,209 @@ ProgramBuilder& ProgramBuilder::startProgram(AtomIndex& index, DefaultUnfoundedC
 }
 
 ProgramBuilder& ProgramBuilder::updateProgram() {
-	if (!incData_)  { incData_ = new Incremental(); }
-	else            {
-		incData_->startVar_   = (uint32)vars_.size();
-		incData_->unfreeze_.clear();
-		for (VarVec::iterator it = incData_->freeze_.begin(), end = incData_->freeze_.end(); it != end; ++it) {
-			atoms_[*it]->setIgnore(false);
-		}
-	}
-	// delete bodies...
+	check_precondition(frozen_ || !incData_, std::logic_error);
+	check_precondition(!atoms_.empty() && "startProgram() not called!", std::logic_error);
+	// delete bodies and clean up atoms
 	disposeProgram(false);
-	// clean up atoms
-	assert(incData_->startAux_ <= atoms_.size());
-	for (VarVec::size_type i = 0; i != incData_->startAux_; ++i) {
-		if (atoms_[i]->eq() && atoms_[i]->eqNode() >= incData_->startAux_) {
-			// atom i is equivalent to some aux atom 
-			std::swap(atoms_[i], atoms_[atoms_[i]->eqNode()]);
-		}
-		VarVec().swap(atoms_[i]->posDep);
-		VarVec().swap(atoms_[i]->negDep);
-		VarVec().swap(atoms_[i]->preds);
+	if (!incData_)  { incData_ = new Incremental(); }
+	incData_->startVar_ = (uint32)vars_.size();
+	incData_->startAtom_= (uint32)atoms_.size();
+	incData_->startAux_ = (uint32)atoms_.size();
+	incData_->unfreeze_.clear();
+	for (VarVec::iterator it = incData_->freeze_.begin(), end = incData_->freeze_.end(); it != end; ++it) {
+		atoms_[*it]->setIgnore(false);
 	}
-	// delete any introduced aux atoms
-	// this is safe because aux atoms are never part of the input program
-	// it is necessary in order to free their ids, i.e. the id of an aux atom
-	// from step I might be needed for a program atom in step I+1
-	for (VarVec::size_type i = incData_->startAux_; i != atoms_.size(); ++i) {
-		delete atoms_[i];
-	}
-	atoms_.erase(atoms_.begin()+incData_->startAux_, atoms_.end());
-	incData_->startAtom_  = (uint32)atoms_.size();
-	incData_->startAux_   = (uint32)atoms_.size();
 	frozen_   = false;
 	atomIndex_->startInit();
 	return *this;
 }
 
+bool ProgramBuilder::endProgram(Solver& solver, bool finalizeSolver, bool backprop) {
+	if (frozen_ == false) {
+		transformExtended();
+		if (normalize_) { normalize(); }
+		stats.atoms = numAtoms() - (startAtom()-1);
+		stats.bodies= numBodies();
+		updateFrozenAtoms(solver);
+		frozen_ = true;
+		if (atoms_[0]->value() == value_true) { return false; }
+		Preprocessor p;
+		p.enableBackprop(backprop);
+		if (!p.preprocess(*this, eqIters_ != 0 ? Preprocessor::full_eq : Preprocessor::no_eq, eqIters_, eqDfs_)) {
+			return false;
+		}
+		if (erMode_ == mode_transform_integ || erMode_ == mode_transform_dynamic) {
+			transformIntegrity(std::min(uint32(15000), uint32(numAtoms())<<1));
+		}
+		vars_.addTo(solver, incData_ ? incData_->startVar_ : 1);
+		atomIndex_->endInit();
+		bodyIndex_.clear();
+		stats.atoms = numAtoms() - (startAtom()-1);
+	}
+	else {
+		if (ufs_.get()) {
+			ufs_ = new DefaultUnfoundedCheck(ufs_->reasonStrategy());
+		}
+		cloneVars(solver);
+	}
+	solver.startAddConstraints();
+	uint32 scc = incData_ ? incData_->startScc_ : 0;
+	CycleChecker c(atoms_, bodies_, ufs_.get() != 0, scc);
+	if (incData_) {
+		for (VarVec::const_iterator it = incData_->unfreeze_.begin(), end = incData_->unfreeze_.end(); it != end; ++it) {
+			atoms_.push_back(atoms_[*it]);
+		}
+	}
+	bool ret = addConstraints(solver, c);
+	if (ufs_.get()) stats.sccs = (c.sccs() - scc);
+	if (ret && !c.sccAtoms().empty()) {
+		uint32 oldNodes = (uint32)ufs_->nodes();
+		if (ufs_.is_owner()) {
+			// Transfer ownership of ufs to solver...
+			solver.addPost(ufs_.release());
+		}
+		// and init the unfounded set checker with new SCCs.
+		ufs_->startInit(solver);
+		ret = ufs_->endInit(c.sccAtoms(), atoms_, bodies_);
+		stats.ufsNodes = (uint32)ufs_->nodes()-oldNodes;
+		if (incData_) incData_->startScc_ = c.sccs();
+	}
+	if (incData_) {
+		atoms_.resize( atoms_.size() - incData_->unfreeze_.size() );
+	}
+	return ret && (!finalizeSolver || solver.endAddConstraints());
+}
+
+MinimizeConstraint* ProgramBuilder::createMinimize(Solver& solver, bool inHeu) {
+	check_precondition(frozen_, std::length_error);
+	if (!minimize_) { return 0; }
+	MinimizeConstraint* m = new MinimizeConstraint();
+	WeightLitVec lits;
+	for (MinimizeRule* r = minimize_; r; r = r->next_) {
+		for (WeightLitVec::iterator it = r->lits_.begin(); it != r->lits_.end(); ++it) {
+			PrgAtomNode* h    = atoms_[it->first.var()];
+			lits.push_back(WeightLiteral(it->first.sign() ? ~h->literal() : h->literal(), it->second));
+		}
+		m->minimize(solver, lits, inHeu);
+		lits.clear();
+	}
+	return m;
+}
+
+void ProgramBuilder::writeProgram(std::ostream& os) {
+	const char* const delimiter = "0";
+	// first write all minimize rules - revert order!
+	PodVector<MinimizeRule*>::type mr;
+	for (MinimizeRule* r = minimize_; r; r = r->next_) {
+		mr.push_back(r);
+	}
+	for (PodVector<MinimizeRule*>::type::reverse_iterator rit = mr.rbegin(); rit != mr.rend(); ++rit) {
+		os << OPTIMIZERULE << " " << 0 << " ";
+		std::stringstream pBody, nBody, weights;
+		VarVec::size_type nbs = 0, pbs =0;
+		MinimizeRule* r = *rit;
+		for (WeightLitVec::iterator it = r->lits_.begin(); it != r->lits_.end(); ++it) {
+			if (atoms_[it->first.var()]->hasVar()) {
+				it->first.sign() ? ++nbs : ++pbs;
+				std::stringstream& body = it->first.sign() ? nBody : pBody;
+				body << it->first.var() << " ";
+				weights << it->second << " ";
+			}
+		}
+		os << pbs+nbs << " " << nbs << " "
+			 << nBody.str() << pBody.str() << weights.str() << "\n";
+	}
+	// write all bodies together with their heads
+	for (BodyList::iterator it = bodies_.begin(); it != bodies_.end(); ++it) {
+		if ( (*it)->hasVar() ) {
+			writeRule(*it, os);
+		}
+	}
+	// write eq-atoms, symbol-table and compute statement
+	std::stringstream bp, bm, symTab;
+	Literal comp;
+	AtomIndex::const_iterator sym = atomIndex_->begin();
+	for (AtomList::size_type i = 1; i < atoms_.size(); ++i) {
+		// write the equivalent atoms
+		if (atoms_[i]->eq()) {
+			os << "1 " << i << " 1 0 " << getEqAtom(Var(i)) << " \n";
+		}
+		if ( atoms_[i]->value() != value_free ) {
+			std::stringstream& str = atoms_[i]->value() == value_false ? bm : bp;
+			str << i << "\n";
+		}
+		if (sym != atomIndex_->end() && Var(i) == sym->first) {
+			if (sym->second.lit != negLit(sentVar) && !sym->second.name.empty()) {
+				symTab << i << " " << sym->second.name << "\n";
+			}
+			++sym;
+		}
+	}
+	os << delimiter << "\n";
+	os << symTab.str();
+	os << delimiter << "\n";
+	os << "B+\n" << bp.str() << "0\n"
+		 << "B-\n" << bm.str() << "0\n1\n";
+}
+/////////////////////////////////////////////////////////////////////////////////////////
+// Program mutating functions
+/////////////////////////////////////////////////////////////////////////////////////////
+#define check_not_frozen() check_precondition(!frozen_ && "Can't update frozen program!", std::logic_error)
+Var ProgramBuilder::newAtom() {
+	check_not_frozen();
+	atoms_.push_back( new PrgAtomNode() );
+	return (Var) atoms_.size() - 1;
+}
+
+ProgramBuilder& ProgramBuilder::setAtomName(Var atomId, const char* name) {
+	check_not_frozen();
+	check_precondition(atomId >= startAtom(), RedefinitionError);
+	resize(atomId);
+	atomIndex_->addUnique(atomId, Atom(name));
+	return *this;
+}
+
+ProgramBuilder& ProgramBuilder::setCompute(Var atomId, bool pos) {
+	PrgAtomNode* a = resize(atomId);
+	ValueRep v     = pos ? value_weak_true : value_false;
+	compute_.push_back(Literal(atomId, v==value_false));
+	if (!a->setValue(v)) {
+		setConflict();
+	}
+	return *this;
+}
+
+ProgramBuilder& ProgramBuilder::freeze(Var atomId) {
+	check_not_frozen();
+	check_precondition(incData_ && "ProgramBuilder::updateProgram() not called!", std::logic_error);
+	PrgAtomNode* a = resize(atomId);
+	if (atomId >= startAtom() && !a->frozen() && a->preds.empty()) {
+		incData_->freeze_.push_back(atomId);
+		a->setFrozen(true);
+	}
+	// else: atom is defined or from a previous step - ignore!
+	return *this;
+}
+
+ProgramBuilder& ProgramBuilder::unfreeze(Var atomId) {
+	check_not_frozen();
+	check_precondition(incData_ && "ProgramBuilder::updateProgram() not called!", std::logic_error);
+	PrgAtomNode* atom = resize(atomId);
+	if (atomId >= startAtom() || atom->frozen()) {
+		incData_->unfreeze_.push_back(atomId);
+	}
+	// else: atom is from a previous step - ignore!
+	return *this;
+}
 
 ProgramBuilder& ProgramBuilder::addRule(PrgRule& r) {
 	// simplify rule, mark literals
+	if (startAtom() > 1) {
+		updateRule(r);
+	}
 	PrgRule::RData rd = r.simplify(ruleState_);
 	if (r.type() != ENDRULE) {     // rule is relevant
+		check_not_frozen();
 		upRules(r.type(), 1);
 		if (handleNatively(r, rd)) { // and can be handled natively
 			addRuleImpl(r, rd);
@@ -903,6 +1084,79 @@ ProgramBuilder& ProgramBuilder::addRule(PrgRule& r) {
 	}
 	return *this;
 }
+#undef check_not_frozen
+/////////////////////////////////////////////////////////////////////////////////////////
+// Query functions
+/////////////////////////////////////////////////////////////////////////////////////////
+Literal ProgramBuilder::getLiteral(Var atomId) const {
+	check_precondition(atomId < atoms_.size(), std::logic_error);
+	return getAtom(getEqAtom(atomId))->literal();
+}
+void ProgramBuilder::getAssumptions(LitVec& out) const {
+	check_precondition(frozen_, std::logic_error);
+	if (incData_) {
+		for (VarVec::const_iterator it = incData_->freeze_.begin(), end = incData_->freeze_.end(); it != end; ++it) {
+			out.push_back( ~getLiteral(*it) );
+		}
+	}
+}
+/////////////////////////////////////////////////////////////////////////////////////////
+// Program definition - private
+/////////////////////////////////////////////////////////////////////////////////////////
+bool ProgramBuilder::inCompute(Literal x) const {
+	return std::find(compute_.begin(), compute_.end(), x) != compute_.end();
+}
+void ProgramBuilder::addRuleImpl(const PrgRule& r, const PrgRule::RData& rd) {
+	if (r.type() != OPTIMIZERULE) {
+		Body b = findOrCreateBody(r, rd);
+		if (b.first->value() == value_false || rd.value == value_false) {
+			// a false body can't define any atoms
+			for (VarVec::iterator it = b.first->heads.begin(), end = b.first->heads.end(); it != end; ++it) {
+				PrgAtomNode* a = atoms_[*it];
+				VarVec::iterator p = std::find(a->preds.begin(), a->preds.end(), b.second);
+				if (p != a->preds.end()) { *p = a->preds.back(); a->preds.pop_back(); }
+			}
+			b.first->heads.clear();
+			b.first->setValue(value_false);
+		}
+		for (VarVec::const_iterator it = r.heads.begin(), end = r.heads.end(); it != end; ++it) {
+			if (b.first->value() != value_false) {
+				PrgAtomNode* a = resize(*it);
+				check_precondition(*it >= startAtom() || a->frozen() || inCompute(negLit(*it)), RedefinitionError);
+				if (r.body.empty()) a->setIgnore(true);
+				if (a->frozen() && a->preds.empty()) {
+					unfreeze(*it);
+				}
+				// Note: b->heads may now contain duplicates. They are removed in PrgBodyNode::buildHeadSet.
+				b.first->heads.push_back((*it));
+				// Similarly, duplicates in atoms_[*it]->preds_ are removed in PrgAtomNode::toConstraint
+				a->preds.push_back(b.second);	
+			}
+			ruleState_.popFromRule(*it);  // clear flag of head atoms
+		}
+		if (rd.value != value_free) { b.first->setValue(rd.value); }
+	}
+	else {
+		check_precondition(r.heads.empty(), std::logic_error);
+		ProgramBuilder::MinimizeRule* mr = new ProgramBuilder::MinimizeRule;
+		for (WeightLitVec::const_iterator it = r.body.begin(), bEnd = r.body.end(); it != bEnd; ++it) {
+			resize(it->first.var());
+			ruleState_.popFromRule(it->first.var());
+		}
+		mr->lits_ = r.body;
+		mr->next_ = minimize_;
+		minimize_ = mr;
+	}
+}
+
+void ProgramBuilder::updateRule(PrgRule& r) {
+	for (WeightLitVec::iterator it = r.body.begin(), end = r.body.end(); it != end; ++it) {
+		Var id = it->first.var();
+		if (id < atoms_.size() && getEqAtom(id) != id) {
+			it->first = Literal(getEqAtom(id), it->first.sign());
+		}
+	}
+}
 
 bool ProgramBuilder::handleNatively(const PrgRule& r, const PrgRule::RData& rd) const {
 	if (erMode_ == mode_native || erMode_ == mode_transform_integ || r.type() == BASICRULE || r.type() == OPTIMIZERULE) {
@@ -928,48 +1182,6 @@ bool ProgramBuilder::handleNatively(const PrgRule& r, const PrgRule::RData& rd)
 	return true;
 }
 
-void ProgramBuilder::addRuleImpl(const PrgRule& r, const PrgRule::RData& rd) {
-	assert(!frozen_ && "Can't add rules to frozen program!");
-	assert(r.type() != ENDRULE);
-	if (r.type() != OPTIMIZERULE) {
-		Body b = findOrCreateBody(r, rd);
-		if (b.first->value() == value_false || rd.value == value_false) {
-			// a false body can't define any atoms
-			for (VarVec::iterator it = b.first->heads.begin(), end = b.first->heads.end(); it != end; ++it) {
-				PrgAtomNode* a = atoms_[*it];
-				VarVec::iterator p = std::find(a->preds.begin(), a->preds.end(), b.second);
-				if (p != a->preds.end()) { *p = a->preds.back(); a->preds.pop_back(); }
-			}
-			b.first->heads.clear();
-			b.first->setValue(value_false);
-		}
-		for (VarVec::const_iterator it = r.heads.begin(), end = r.heads.end(); it != end; ++it) {
-			if (b.first->value() != value_false) {
-				PrgAtomNode* a = resize(*it);
-				assert( (!a->hasVar() || (incData_ && incData_->frozen(*it))) && "Cannot extend definition of atom!");
-				if (r.body.empty()) a->setIgnore(true);
-				// Note: b->heads may now contain duplicates. They are removed in PrgBodyNode::buildHeadSet.
-				b.first->heads.push_back((*it));
-				// Similarly, duplicates in atoms_[*it]->preds_ are removed in PrgAtomNode::toConstraint
-				a->preds.push_back(b.second);
-			}
-			ruleState_.popFromRule(*it);  // clear flag of head atoms
-		}
-		if (rd.value != value_free) { b.first->setValue(rd.value); }
-	}
-	else {
-		assert(r.heads.empty());
-		ProgramBuilder::MinimizeRule* mr = new ProgramBuilder::MinimizeRule;
-		for (WeightLitVec::const_iterator it = r.body.begin(), bEnd = r.body.end(); it != bEnd; ++it) {
-			resize(it->first.var());
-			ruleState_.popFromRule(it->first.var());
-		}
-		mr->lits_ = r.body;
-		mr->next_ = minimize_;
-		minimize_ = mr;
-	}
-}
-
 ProgramBuilder::Body ProgramBuilder::findOrCreateBody(const PrgRule& r, const PrgRule::RData& rd) {
 	// Note: We don't search for an existing body node for weight rules because
 	// for those checking for equivalence can't be done using the "mark-and-test"-approach alone.
@@ -1158,99 +1370,56 @@ void ProgramBuilder::normalize() {
 	}	
 }
 
-bool ProgramBuilder::endProgram(Solver& solver, bool finalizeSolver, bool backprop) {
-	if (frozen_ == false) {
-		transformExtended();
-		if (normalize_) { normalize(); }
-		stats.atoms = numAtoms() - (startAtom()-1);
-		stats.bodies= numBodies();
-		updateFrozenAtoms(solver);
-		frozen_ = true;
-		if (atoms_[0]->value() == value_true) { return false; }
-		Preprocessor p;
-		p.enableBackprop(backprop);
-		if (!p.preprocess(*this, eqIters_ != 0 ? Preprocessor::full_eq : Preprocessor::no_eq, eqIters_, eqDfs_)) {
-			return false;
-		}
-		if (erMode_ == mode_transform_integ || erMode_ == mode_transform_dynamic) {
-			transformIntegrity(std::min(uint32(15000), uint32(numAtoms())<<1));
-		}
-		vars_.addTo(solver, incData_ ? incData_->startVar_ : 1);
-		atomIndex_->endInit();
-		bodyIndex_.clear();
-		stats.atoms = numAtoms() - (startAtom()-1);
-	}
-	else {
-		if (ufs_.get()) {
-			ufs_ = new DefaultUnfoundedCheck(ufs_->reasonStrategy());
-		}
-		cloneVars(solver);
-	}
-	solver.startAddConstraints();
-	CycleChecker c(atoms_, bodies_, ufs_.get() != 0);
-	if (incData_) {
-		for (VarVec::const_iterator it = incData_->unfreeze_.begin(), end = incData_->unfreeze_.end(); it != end; ++it) {
-			atoms_.push_back(atoms_[*it]);
-		}
-	}
-	bool ret = addConstraints(solver, c);
-	if (ufs_.get()) stats.sccs = c.sccs();
-	if (ret && c.hasCycles()) {
-		uint32 oldNodes = (uint32)ufs_->nodes();
-		if (ufs_.is_owner()) {
-			// Transfer ownership of ufs to solver...
-			solver.addPost(ufs_.release());
-		}
-		// and init the unfounded set checker with new SCCs.
-		ufs_->startInit(solver);
-		ret = ufs_->endInit(c.sccAtoms(), atoms_, bodies_);
-		stats.ufsNodes = (uint32)ufs_->nodes()-oldNodes;
-	}
-	if (incData_) atoms_.resize( atoms_.size() - incData_->unfreeze_.size() );
-	return ret && (!finalizeSolver || solver.endAddConstraints());
-}
-
 void ProgramBuilder::updateFrozenAtoms(const Solver& solver) {
 	if (incData_ != 0) {
 		// update truth values of atoms from previous iterations
 		for (uint32 i = 1; i != incData_->startAtom_; ++i) {
-			ValueRep v = solver.value( atoms_[i]->var() );
-			if (v != value_free) {
-				atoms_[i]->setValue(v == trueValue(atoms_[i]->literal()) ? value_true : value_false);
+			ValueRep v;
+			if (atoms_[i]->hasVar() && (v = solver.value(atoms_[i]->var())) != value_free) {
+				if (v == trueValue(atoms_[i]->literal())) {
+					// set to strong true only if we are sure that atom has a valid support
+					v = atoms_[i]->frozen() || atoms_[i]->value() != value_true ? value_weak_true : value_true;
+				}
+				else {
+					v = value_false;
+				}
+				if (v != atoms_[i]->value()) {
+					atoms_[i]->setValue(v);
+				}
 			}
 		}
-		ExtendedRuleMode oldMode = erMode_;
-		erMode_ = mode_native;
+		// remove protection of frozen atoms
+		VarVec::iterator j = incData_->unfreeze_.begin();
 		for (VarVec::iterator it = incData_->unfreeze_.begin(), end = incData_->unfreeze_.end(); it != end; ++it) {
 			PrgAtomNode* a = atoms_[*it];
-			assert(a->hasVar() && "Error: Can't unfreeze atom that is not frozen");
-			// Mark literal of a so that we can remove a from freeze-list
-			Literal x = a->literal();
-			x.watch();
-			a->setLiteral( x );
+			if (a->frozen() && *it < startAtom()) {
+				a->resetSccFlags();
+				*j++ = *it;
+			}
+			a->setFrozen(false);
 		}
-		VarVec::iterator j = incData_->freeze_.begin();
+		incData_->unfreeze_.erase(j, incData_->unfreeze_.end());
+		// add protection for atoms still frozen
+		j = incData_->freeze_.begin();
+		Body emptyBody(0,0);
 		for (VarVec::iterator it = j, end = incData_->freeze_.end(); it != end; ++it) {
 			PrgAtomNode* a = atoms_[*it];
-			if (a->literal().watched()) { // atom is no longer frozen
-				Literal x = a->literal();
-				x.clearWatch();
-				a->setLiteral(x);
-				a->resetSccFlags();
-			}
-			else {
+			if (a->frozen()) {
 				assert(a->preds.empty() && "Can't freeze defined atom!");
+				if (emptyBody.first == 0) { 
+					rule_.clear(); rule_.setType(CHOICERULE);
+					PrgRule::RData rd = {0,0,0,0};
+					emptyBody = findOrCreateBody(rule_, rd); 
+				}
 				// Make atom a choice.
 				// This way, no special handling during preprocessing/nogood creation is necessary
-				// Note: The choice is removed on next call to updateProgram!
-				startRule(CHOICERULE).addHead(*it).endRule();
-				// Note: Don't count this rule - it is only an implementation detail
-				upRules(CHOICERULE, -1);
+				emptyBody.first->heads.push_back(*it);
+				a->preds.push_back(emptyBody.second);	
+				a->setIgnore(true);
 				*j++ = *it;
 			}
 		}
 		incData_->freeze_.erase(j, incData_->freeze_.end());
-		erMode_ = oldMode;
 	}
 }
 
@@ -1266,7 +1435,7 @@ bool ProgramBuilder::mergeEqAtoms(Var a, Var root) {
 		setConflict();
 		return false;
 	}
-	assert(at->value() == r->value());
+	assert(at->value() == r->value() || (r->value() == value_true && at->value() == value_weak_true));
 	at->setEq(root);
 	incEqs(Var_t::atom_var);
 	return true;
@@ -1285,6 +1454,8 @@ bool ProgramBuilder::addConstraints(Solver& s, CycleChecker& c) {
 	}
 	const bool freezeAtoms = incData_ && s.strategies().satPrePro.get() != 0;
 	uint32 start           = startAtom();
+	check_precondition(atomIndex_->curBegin() == atomIndex_->end() || start <= atomIndex_->curBegin()->first,
+		std::logic_error);
 	AtomIndex::const_iterator sym = atomIndex_->lower_bound(atomIndex_->curBegin(), start);
 	for (AtomList::const_iterator it = atoms_.begin()+start; it != atoms_.end(); ++it) {
 		if ( !(*it)->toConstraint(s, gc, *this) ) { return false; }
@@ -1301,20 +1472,6 @@ bool ProgramBuilder::addConstraints(Solver& s, CycleChecker& c) {
 	return true;
 }
 
-MinimizeConstraint* ProgramBuilder::createMinimize(Solver& solver, bool inHeu) {
-	if (!minimize_) { return 0; }
-	MinimizeConstraint* m = new MinimizeConstraint();
-	WeightLitVec lits;
-	for (MinimizeRule* r = minimize_; r; r = r->next_) {
-		for (WeightLitVec::iterator it = r->lits_.begin(); it != r->lits_.end(); ++it) {
-			PrgAtomNode* h    = atoms_[it->first.var()];
-			lits.push_back(WeightLiteral(it->first.sign() ? ~h->literal() : h->literal(), it->second));
-		}
-		m->minimize(solver, lits, inHeu);
-		lits.clear();
-	}
-	return m;
-}
 
 // exclude vars contained in minimize statements from var elimination
 void ProgramBuilder::freezeMinimize(Solver& solver) {
@@ -1335,12 +1492,6 @@ void ProgramBuilder::setConflict() {
 	atoms_[0]->setValue(value_true);
 }
 
-Var ProgramBuilder::newAtom() {
-	assert(!frozen_ && "Can't add atom to frozen program!");
-	atoms_.push_back( new PrgAtomNode() );
-	return (Var) atoms_.size() - 1;
-}
-
 PrgAtomNode* ProgramBuilder::resize(Var atomId) {
 	assert(atomId != varMax && atomId > 0);
 	while (atoms_.size() <= AtomList::size_type(atomId)) {
@@ -1366,62 +1517,6 @@ void ProgramBuilder::cloneVars(Solver& solver) {
 	}
 }
 
-void ProgramBuilder::writeProgram(std::ostream& os) {
-	const char* const delimiter = "0";
-	// first write all minimize rules - revert order!
-	PodVector<MinimizeRule*>::type mr;
-	for (MinimizeRule* r = minimize_; r; r = r->next_) {
-		mr.push_back(r);
-	}
-	for (PodVector<MinimizeRule*>::type::reverse_iterator rit = mr.rbegin(); rit != mr.rend(); ++rit) {
-		os << OPTIMIZERULE << " " << 0 << " ";
-		std::stringstream pBody, nBody, weights;
-		VarVec::size_type nbs = 0, pbs =0;
-		MinimizeRule* r = *rit;
-		for (WeightLitVec::iterator it = r->lits_.begin(); it != r->lits_.end(); ++it) {
-			if (atoms_[it->first.var()]->hasVar()) {
-				it->first.sign() ? ++nbs : ++pbs;
-				std::stringstream& body = it->first.sign() ? nBody : pBody;
-				body << it->first.var() << " ";
-				weights << it->second << " ";
-			}
-		}
-		os << pbs+nbs << " " << nbs << " "
-			 << nBody.str() << pBody.str() << weights.str() << "\n";
-	}
-	// write all bodies together with their heads
-	for (BodyList::iterator it = bodies_.begin(); it != bodies_.end(); ++it) {
-		if ( (*it)->hasVar() ) {
-			writeRule(*it, os);
-		}
-	}
-	// write eq-atoms, symbol-table and compute statement
-	std::stringstream bp, bm, symTab;
-	Literal comp;
-	AtomIndex::const_iterator sym = atomIndex_->begin();
-	for (AtomList::size_type i = 1; i < atoms_.size(); ++i) {
-		// write the equivalent atoms
-		if (atoms_[i]->eq()) {
-			os << "1 " << i << " 1 0 " << getEqAtom(Var(i)) << " \n";
-		}
-		if ( atoms_[i]->value() != value_free ) {
-			std::stringstream& str = atoms_[i]->value() == value_false ? bm : bp;
-			str << i << "\n";
-		}
-		if (sym != atomIndex_->end() && Var(i) == sym->first) {
-			if (sym->second.lit != negLit(sentVar) && !sym->second.name.empty()) {
-				symTab << i << " " << sym->second.name << "\n";
-			}
-			++sym;
-		}
-	}
-	os << delimiter << "\n";
-	os << symTab.str();
-	os << delimiter << "\n";
-	os << "B+\n" << bp.str() << "0\n"
-		 << "B-\n" << bm.str() << "0\n1\n";
-}
-
 void ProgramBuilder::writeRule(PrgBodyNode* b, std::ostream& os) {
 	VarVec::size_type nbs = 0, pbs = 0;
 	std::stringstream body;
