@@ -45,7 +45,9 @@ bool WeightConstraint::newWeightConstraint(Solver& s, Literal con, WeightLitVec&
 				return false;
 			}
 		}
-		if (i == lits.size()) return true;
+		if (i == lits.size() || bound <= 0) {
+			return true;
+		}
 		lits.erase(lits.begin(), lits.begin()+i);
 		if (lits.front().second == lits.back().second && lits.front().second != 1) {
 			// disguised cardinality constraint
@@ -118,6 +120,9 @@ weight_t WeightConstraint::canonicalize(Solver& s, WeightLitVec& lits, weight_t&
 		if (uint32(lits[i].second) < minW) { minW = lits[i].second; }
 		sumW += lits[i].second;
 	}
+	if (bound < 0) { 
+		bound = 0;
+	}
 	// Step 3: sort by decreasing weight
 	if (maxW != minW) {
 		std::stable_sort(lits.begin(), lits.end(), compose22(
@@ -162,7 +167,7 @@ WeightConstraint::WeightConstraint(Solver& s, Literal con, const WeightLitVec& l
 		s.setFrozen(lits_[idx].var(), true);    // exempt from variable elimination
 	}
 	// Initialize heuristic with literals (no weights) in constraint.
-	s.strategies().heuristic->newConstraint(s, &lits_[undoStart()], size_, Constraint_t::native_constraint);
+	s.strategies().heuristic->newConstraint(s, &lits_[undoStart()], size_, Constraint_t::static_constraint);
 	if (active_ == NOT_ACTIVE) {
 		addWatch(s, 0, FTB_BFB);                // watch con in both phases
 		addWatch(s, 0, FFB_BTB);                // in order to allow for backpropagation
@@ -275,7 +280,7 @@ Constraint::PropResult WeightConstraint::propagate(const Literal&, uint32& d, So
 // Builds the reason for p from the undo stack of this constraint.
 // The reason will only contain literals that were processed by the 
 // active sub-constraint.
-void WeightConstraint::reason(const Literal& p, LitVec& r) {
+ConstraintType WeightConstraint::reason(const Literal& p, LitVec& r) {
 	assert(active_ != NOT_ACTIVE);
 	Literal x;
 	for (uint32 i = undoStart(); i <= undo_; ++i) {
@@ -292,6 +297,7 @@ void WeightConstraint::reason(const Literal& p, LitVec& r) {
 			}
 		}
 	}
+	return WeightConstraint::type();
 }
 
 // undoes processed assignments 
@@ -405,6 +411,7 @@ void MinimizeConstraint::minimize(Solver& s, const WeightLitVec& literals, bool
 			++j;
 		}
 	}
+	nr->lits_.erase(j, nr->lits_.end());
 }
 
 bool MinimizeConstraint::simplify(Solver&, bool) {
@@ -488,7 +495,7 @@ bool MinimizeConstraint::backpropagate(Solver& s, MinRule* r) {
 	return r->sum_ <= r->opt_;
 }
 
-void MinimizeConstraint::reason(const Literal& p, LitVec& r) {
+ConstraintType MinimizeConstraint::reason(const Literal& p, LitVec& r) {
 	uint32 activeRule = 0;
 	if (minRules_.size() > 1) {
 		UndoList::iterator it = undoList_.end() - 1;
@@ -500,6 +507,7 @@ void MinimizeConstraint::reason(const Literal& p, LitVec& r) {
 			r.push_back(undoList_[i].lit_);
 		}
 	}
+	return MinimizeConstraint::type();
 }
 
 uint64 MinimizeConstraint::models() const {
