@@ -64,7 +64,7 @@ public:
 	void   endInit(Solver& s) {
 		vars_.clear();
 		for (Var i = 1; i <= s.numVars(); ++i) {
-			if (s.value( i ) == value_free) {
+			if (s.finalValue( i ) == value_free) {
 				vars_.push_back(i);
 			}
 		}
@@ -267,7 +267,7 @@ void Solver::startAddConstraints() {
 
 bool Solver::endAddConstraints() {
 	Antecedent::checkPlatformAssumptions();
-	if (strategy_.satPrePro.get() != 0) {
+	if (strategy_.satPrePro.get() != 0 && decisionLevel() == 0) {
 		SolverStrategies::SatPrePro temp(strategy_.satPrePro.release());
 		bool r = temp->preprocess();
 		strategy_.satPrePro.reset(temp.release());
@@ -329,10 +329,7 @@ void Solver::eliminate(Var v, bool elim) {
 }
 
 bool Solver::addUnary(Literal p) {
-	if (decisionLevel() != 0) {
-		impliedLits_.push_back(ImpliedLiteral(p, 0, posLit(0)));
-	}
-	return force(p, posLit(0));
+	return addNewImplication(p, 0, Antecedent(posLit(0)));
 }
 
 bool Solver::addBinary(Literal p, Literal q) {
@@ -357,13 +354,14 @@ void Solver::setStopConflict() {
 		// we use the nogood {FALSE} to represent the unrecoverable conflict -
 		// note that {FALSE} can otherwise never be a violated nogood because
 		// TRUE is always true in every solver
-		conflict_.push_back(negLit(0));
+		conflict_.lits.push_back(negLit(0));
 	}
 	// artificially increase root level -
 	// this way, the solver is prevented from resolving the conflict
 	setRootLevel(decisionLevel());
+	conflict_.level = decisionLevel();
 }
-bool Solver::hasStopConflict() const { return conflict_[0] == negLit(0); }
+bool Solver::hasStopConflict() const { return conflict_.lits[0] == negLit(0); }
 bool Solver::clearAssumptions()  {
 	rootLevel_ = btLevel_ = 0;
 	undoUntil(0);
@@ -383,7 +381,7 @@ bool Solver::integrateClause(const LitVec& lits, bool addSat) {
 	assert(!hasConflict()); // Precondition
 	// Add clause to clause creator -
 	// determine state of clause
-	ClauseCreator nc(this); nc.start(Constraint_t::learnt_conflict);
+	ClauseCreator nc(this); nc.start(Constraint_t::learnt_other);
 	for (LitVec::size_type i = 0; i != lits.size(); ++i) {  nc.add(lits[i]);  }
 	
 	if (nc.end(addSat)) { return true; }
@@ -394,19 +392,20 @@ bool Solver::integrateClause(const LitVec& lits, bool addSat) {
 		// nc is already an asserting clause - i.e.
 		// resolveConflict() would generate the exact same clause.
 		undoUntil(nc.implicationLevel());
-		ClauseCreator::addClause(*this, nc.type(), nc.lits(), nc.sw());
-		return true;
+		return ClauseCreator::addClause(*this, nc.type(), nc.lits(), nc.sw());
 	}
 	undoUntil(nc.conflictLevel());
-	if (decisionLevel() == rootLevel()) {
-		// remember the clause, because resolveConflict() will return
-		// false and won't learn anything from the conflict.
+	if (decisionLevel() == rootLevel() || decisionLevel() > nc.conflictLevel()) {
+		// remember the clause, because resolveConflict() won't learn anything 
+		// from the conflict - either because root level conflicts are not resolved
+		// or because the current backtracking level prevents learning.
 		return ClauseCreator::addClause(*this, nc.type(), nc.lits(), nc.sw());
 	}
 	// store conflict
 	conflict_.clear();
+	conflict_.level = nc.conflictLevel();
 	for (LitVec::size_type i = 0; i != nc.size(); ++i) {
-		conflict_.push_back( ~nc[i] );
+		conflict_.lits.push_back( ~nc[i] );
 	}
 	return false;
 }
@@ -575,9 +574,10 @@ bool Solver::force(const Literal& p, const Antecedent& c) {
 		return true;
 	}
 	// else: conflict
-	conflict_.push_back(~p);
+	conflict_.level = decisionLevel();
+	conflict_.lits.push_back(~p);
 	if (strategy_.search != SolverStrategies::no_learning && !c.isNull()) {
-		c.reason(p, conflict_);
+		stats.solve.updateAntes(c.reason(p, conflict_.lits));
 	}
 	return false;
 }
@@ -672,18 +672,36 @@ bool Solver::test(Literal p, Constraint* c) {
 	return false;
 }
 
+void Solver::setConflict(LitVec& cfl, bool computeLevel) {
+	conflict_.clear(); 
+	uint32 maxDL = !computeLevel ? decisionLevel() : 0;
+	for (LitVec::const_iterator it = cfl.begin(), end = cfl.end(); it != end && maxDL != decisionLevel(); ++it) {
+		if (level(it->var()) > maxDL) {
+			maxDL = level(it->var());
+		}
+	}
+	conflict_.level = maxDL;
+	conflict_.lits.swap(cfl);
+}
+
 bool Solver::resolveConflict() {
-	assert(hasConflict());
+	assert(hasConflict() && conflict_.level <= decisionLevel());
 	++stats.solve.conflicts;
+	undoUntil(conflict_.level);
 	if (decisionLevel() > rootLevel_) {
 		if (decisionLevel() != btLevel_ && strategy_.search != SolverStrategies::no_learning) {
 			uint32 sw;
 			uint32 uipLevel = analyzeConflict(sw);
 			stats.solve.updateJumps(decisionLevel(), uipLevel, btLevel_);
 			undoUntil( uipLevel );
+			assert(!hasConflict());
 			return ClauseCreator::addClause(*this, Constraint_t::learnt_conflict, cc_, sw);
 		}
 		else {
+			if (btLevel_ > conflict_.level) {
+				setBacktrackLevel(conflict_.level);
+				undoUntil(conflict_.level);
+			}
 			return backtrack();
 		}
 	}
@@ -694,27 +712,33 @@ bool Solver::resolveConflict() {
 }
 
 bool Solver::backtrack() {
-	if (decisionLevel() == rootLevel_) return false;
-	Literal lastChoiceInverted = ~decision(decisionLevel());
-	btLevel_ = decisionLevel() - 1;
-	undoUntil(btLevel_);
-	bool r = force(lastChoiceInverted, 0);
-	ImpliedLits::size_type j = 0;
-	for (ImpliedLits::size_type i = 0; i < impliedLits_.size(); ++i) {
-		r = r && force(impliedLits_[i].lit, impliedLits_[i].ante);
-		if (impliedLits_[i].level != btLevel_) {
-			impliedLits_[j++] = impliedLits_[i];
+	do {
+		if (decisionLevel() == rootLevel_) return false;
+		Literal lastChoiceInverted = ~decision(decisionLevel());
+		btLevel_ = decisionLevel() - 1;
+		undoUntil(btLevel_);
+		if (!hasConflict() && force(lastChoiceInverted, 0)) {
+			ImpliedLits::size_type j = 0;
+			for (ImpliedLits::size_type i = 0; i < impliedLits_.size(); ++i) {
+				if (impliedLits_[i].level <= btLevel_) {
+					if (!hasConflict()) { force(impliedLits_[i].lit, impliedLits_[i].ante); }
+					impliedLits_[j++] = impliedLits_[i];
+				}
+			}
+			if (decisionLevel() == 0) { j = 0; }
+			impliedLits_.erase(impliedLits_.begin()+j, impliedLits_.end());
 		}
-	}
-	impliedLits_.erase(impliedLits_.begin()+j, impliedLits_.end());
-	return r || backtrack();
+	} while (hasConflict());
+	return true;
 }
 
 void Solver::undoUntil(uint32 level) {
 	assert(btLevel_ >= rootLevel_);
 	level = std::max( level, btLevel_ );
 	if (level >= decisionLevel()) return;
-	conflict_.clear();
+	if (level < conflict_.level) {
+		conflict_.clear();
+	}
 	strategy_.heuristic->undoUntil( *this, levels_[level].first);
 	bool sp = strategy_.saveProgress > 0 && (decisionLevel() - level) > (uint32)strategy_.saveProgress;
 	undoLevel(false);
@@ -792,10 +816,11 @@ uint32 Solver::analyzeConflict(uint32& secondWatch) {
 	uint32 abstr    = 0;        // abstraction of DLs in cc_
 	Literal p;                  // literal to be resolved out next
 	cc_.assign(1, p);           // will later be replaced with asserting literal
-	strategy_.heuristic->updateReason(*this, conflict_, p);
+	LitVec& rhs    = conflict_.lits;
+	strategy_.heuristic->updateReason(*this, rhs, p);
 	for (;;) {
-		for (LitVec::size_type i = 0; i != conflict_.size(); ++i) {
-			Literal& q = conflict_[i];
+		for (LitVec::size_type i = 0; i != rhs.size(); ++i) {
+			Literal& q = rhs[i];
 			if (!seen(q.var())) {
 				assert(isTrue(q) && "Invalid literal in reason set!");
 				uint32 cl = level(q.var());
@@ -816,12 +841,12 @@ uint32 Solver::analyzeConflict(uint32& secondWatch) {
 		}
 		p = assign_.last();
 		clearSeen(p.var());
-		conflict_.clear();
+		rhs.clear();
 		if (--onLevel == 0) {
 			break;
 		}
-		reason(p).reason(p, conflict_);
-		strategy_.heuristic->updateReason(*this, conflict_, p);
+		extractReason(p, rhs);
+		strategy_.heuristic->updateReason(*this, rhs, p);
 	}
 	cc_[0] = ~p; // store the 1-UIP
 	assert( decisionLevel() == level(p.var()));
@@ -841,6 +866,46 @@ uint32 Solver::analyzeConflict(uint32& secondWatch) {
 	return assertLevel;
 }
 
+void Solver::analyzeRootConflict(Literal p) {
+	uint32 marked = 0;  // number of literals to resolve out
+	uint32 tPos   = 0;  // current position in trail
+	cc_.clear();        // will later hold the final result
+	LitVec rhs    = conflict_.lits;
+	if (!isSentinel(p)) {
+		assert(isTrue(p));
+		cc_.push_back(p);
+		rhs.push_back(p);
+	}
+	if (decisionLevel() == 0) return;
+	for (tPos = assign_.trail.size();;) {
+		// process current rhs
+		for (LitVec::size_type i = 0; i != rhs.size(); ++i) {
+			Literal& q = rhs[i];
+			if (!seen(q.var())) {
+				assert(isTrue(q) && "Invalid literal in reason set!");
+				assert(level(q.var()) > 0 && "Top-Level implication not marked!");
+				markSeen(q.var());
+				++marked;
+			}
+		}
+		rhs.clear();
+		if (marked == 0) { break; }
+		// search for the last assigned literal that needs to be analyzed...
+		while (!seen(assign_.trail[--tPos].var())) {
+			;
+		}	
+		p = assign_.trail[tPos];
+		clearSeen(p.var());
+		--marked;
+		if (reason(p).isNull()) {
+			cc_.push_back(~p);
+		}
+		else {
+			extractReason(p, rhs);
+		}
+	}
+}
+
 void Solver::minimizeConflictClause(uint32 abstr) {
 	uint32 m = strategy_.cflMinAntes;
 	LitVec::size_type t = assign_.trail.size();
@@ -862,10 +927,11 @@ void Solver::minimizeConflictClause(uint32 abstr) {
 }
 
 bool Solver::minimizeLitRedundant(Literal p, uint32 abstr) {
+	LitVec& temp = conflict_.lits;
 	if (!strategy_.strengthenRecursive) {
-		conflict_.clear(); reason(p).reason(p, conflict_);
-		for (LitVec::size_type i = 0; i != conflict_.size(); ++i) {
-			if (!seen(conflict_[i].var())) {
+		temp.clear(); extractReason(p, temp);
+		for (LitVec::size_type i = 0; i != temp.size(); ++i) {
+			if (!seen(temp[i].var())) {
 				return false;
 			}
 		}
@@ -877,9 +943,9 @@ bool Solver::minimizeLitRedundant(Literal p, uint32 abstr) {
 	LitVec::size_type start = assign_.trail.size();
 	for (LitVec::size_type f = start;;) {
 		conflict_.clear();
-		reason(p).reason(p, conflict_);
-		for (LitVec::size_type i = 0; i != conflict_.size(); ++i) {
-			p = conflict_[i];
+		extractReason(p, temp);
+		for (LitVec::size_type i = 0; i != temp.size(); ++i) {
+			p = temp[i];
 			if (!seen(p.var())) {
 				if (!reason(p).isNull() && ((1<<(level(p.var())&31)) & abstr) != 0) {
 					markSeen(p.var());
