@@ -471,7 +471,7 @@ Constraint::PropResult DefaultUnfoundedCheck::propagate(const Literal& p, uint32
 	return PropResult(true, true);  // always keep the watch
 }
 // only used if strategy_ == only_reason or conflict
-void DefaultUnfoundedCheck::reason(const Literal& p, LitVec& r) {
+ConstraintType DefaultUnfoundedCheck::reason(const Literal& p, LitVec& r) {
 	LitVec::const_iterator it, end;
 	if (!activeClause_->empty() && (*activeClause_)[0] == p) {
 		it  = activeClause_->lits().begin()+1;
@@ -483,6 +483,7 @@ void DefaultUnfoundedCheck::reason(const Literal& p, LitVec& r) {
 		end = reasons_[p.var()-1].end();
 	}
 	for (; it != end; ++it) r.push_back( ~*it );
+	return Constraint_t::static_constraint;
 }
 
 /////////////////////////////////////////////////////////////////////////////////////////
